import os

from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph import GraphicsLayoutWidget, ColorButton

from ..plot_widgets.ImgWidget import IntegrationBatchWidget
from .CustomWidgets import MouseCurrentAndClickedWidget
from ..CustomWidgets import (
    FlatButton,
    CheckableFlatButton,
    HorizontalSpacerItem,
    VerticalSpacerItem,
    LabelAlignRight,
    LabelExpandable,
)

from . import CLICKED_COLOR
from ... import icons_path

try:
    from ..plot_widgets.SurfaceWidget import SurfaceWidget

    open_gl = True
except ModuleNotFoundError:
    open_gl = False


class BatchWidget(QtWidgets.QWidget):
    """
    Class describe a widget for batch integration
    """

    def __init__(self, parent=None):
        super(BatchWidget, self).__init__(parent)

        self.frame = QtWidgets.QWidget()
        self.frame.setObjectName("batch_frame")

        self._frame_layout = QtWidgets.QVBoxLayout()
        self._top_layout = QtWidgets.QHBoxLayout()
        self._central_layout = QtWidgets.QHBoxLayout()

        self.black_container = QtWidgets.QWidget()
        self.black_container.setObjectName("black_container")
        self._black_container_layout = QtWidgets.QVBoxLayout()

        # top
        self.file_control_widget = BatchFileControlWidget()
        self.mode_widget = BatchModeWidget()

        # central
        self.file_view_widget = BatchFileViewWidget()
        self.stack_plot_widget = BatchStackWidget()
        if open_gl:
            self.surface_widget = BatchSurfaceWidget()
        self.options_widget = BatchOptionsWidget()

        # bottom
        self.control_widget = BatchControlWidget()
        self.position_widget = BatchImgPositionWidget()

        self.create_layout()
        self.style_widgets()

        self.setLayout(self._layout)
        self.setWindowFlags(QtCore.Qt.Tool)
        self.setAttribute(QtCore.Qt.WA_MacAlwaysShowToolWindow)
        self.setWindowTitle("Batch widget")

        self.activate_files_view()

    def create_layout(self):
        self._top_layout.addWidget(self.file_control_widget)
        self._top_layout.addWidget(self.mode_widget)

        self._central_layout.addWidget(self.file_view_widget)
        self._central_layout.addWidget(self.stack_plot_widget)
        if open_gl:
            self._central_layout.addWidget(self.surface_widget)
        self._central_layout.addWidget(self.options_widget)

        self._frame_layout.addLayout(self._top_layout)

        self._black_container_layout.addLayout(self._central_layout)
        self._black_container_layout.addWidget(self.control_widget)
        self.black_container.setLayout(self._black_container_layout)

        self._frame_layout.addWidget(self.black_container)
        self._frame_layout.addWidget(self.position_widget)
        
        self._frame_layout.setStretchFactor(self.black_container, 1)
        self._frame_layout.setStretchFactor(self.position_widget, 0)
        self._frame_layout.setStretchFactor(self._top_layout, 0)

        self.frame.setLayout(self._frame_layout)
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.addWidget(self.frame)

    def style_widgets(self):
        zero_spacing_layouts = [
            self._frame_layout,
            self._top_layout,
            self._layout,
            self._central_layout,
            self._black_container_layout,
        ]

        for layout in zero_spacing_layouts:
            layout.setContentsMargins(0, 0, 0, 0)
            layout.setSpacing(0)

        self.setStyleSheet(
            """
            #black_container {
                background: black;
            }
            """
        )

    def sizeHint(self):
        return QtCore.QSize(800, 600)

    def activate_files_view(self):
        self.mode_widget.view_f_btn.setChecked(True)

        self.file_view_widget.show()
        if open_gl:
            self.surface_widget.hide()
        self.stack_plot_widget.hide()
        self.options_widget.hide()

        self.position_widget.step_raw_widget.show()
        self.position_widget.step_series_widget.hide()
        self.control_widget.waterfall_btn.hide()
        self.control_widget.phases_btn.hide()
        self.control_widget.autoscale_btn.hide()
        self.control_widget.normalize_btn.hide()
        self.control_widget.integrate_btn.show()

    def activate_stack_plot(self):
        self.position_widget.step_raw_widget.hide()
        self.position_widget.step_series_widget.show()
        self.mode_widget.view_2d_btn.setChecked(True)

        self.file_view_widget.hide()
        self.stack_plot_widget.show()
        self.options_widget.show()
        if open_gl:
            self.surface_widget.hide()
        self.control_widget.waterfall_btn.show()
        self.control_widget.phases_btn.show()
        self.control_widget.autoscale_btn.show()
        self.control_widget.normalize_btn.show()
        self.control_widget.integrate_btn.hide()

    def activate_surface_view(self):
        self.position_widget.step_raw_widget.hide()
        self.position_widget.step_series_widget.show()

        self.mode_widget.view_3d_btn.setChecked(True)

        y = self.position_widget.step_series_widget.slider.value()
        self.surface_widget.surface_view.g_translate = y

        self.file_view_widget.hide()
        self.stack_plot_widget.hide()
        self.options_widget.show()
        self.surface_widget.show()

        self.control_widget.waterfall_btn.hide()
        self.control_widget.phases_btn.hide()
        self.control_widget.autoscale_btn.hide()
        self.control_widget.normalize_btn.hide()
        self.control_widget.integrate_btn.hide()

    def raise_widget(self):
        self.show()
        self.setWindowState(
            self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive
        )
        self.activateWindow()
        self.raise_()


class BatchFileViewWidget(QtWidgets.QWidget):
    """
    Widget to show raw files, calibration and mask files

    Widget contains:
        QLine: calibration file
        QLine: mask file
        QTreeView: raw files
    """

    iteration_name = ""

    def __init__(self):
        super(BatchFileViewWidget, self).__init__()

        self._layout = QtWidgets.QVBoxLayout()
        self._file_lbl_widget = QtWidgets.QWidget()
        self._file_lbl_layout = QtWidgets.QGridLayout()

        self.cal_file_lbl = LabelExpandable("undefined")
        self.mask_file_lbl = LabelExpandable("undefined")

        self.treeView = QtWidgets.QTreeView()
        self.treeView.setObjectName("treeView")

        self.tree_model = QtGui.QStandardItemModel()
        self.treeView.setModel(self.tree_model)
        self.treeView.setColumnWidth(0, 350)

        self.style_widgets()
        self.create_layout()

    def create_layout(self):
        self._file_lbl_layout.addWidget(LabelAlignRight("Calibration File:"), 0, 0)
        self._file_lbl_layout.addWidget(self.cal_file_lbl, 0, 1)
        self._file_lbl_layout.addItem(HorizontalSpacerItem(), 0, 2)

        self._file_lbl_layout.addWidget(LabelAlignRight("Mask File:"), 1, 0)
        self._file_lbl_layout.addWidget(self.mask_file_lbl, 1, 1)

        self._file_lbl_widget.setLayout(self._file_lbl_layout)
        self._layout.addWidget(self._file_lbl_widget)
        self._layout.addWidget(self.treeView)
        self.setLayout(self._layout)

    def style_widgets(self):
        self._file_lbl_layout.setContentsMargins(5, 3, 5, 2)
        self._file_lbl_layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setStyleSheet(
            """
            #treeView {
                background: black;
                color: yellow;
            }
        """
        )

    def set_raw_files(self, files, images):
        self.tree_model.clear()
        self.tree_model.setColumnCount(2)
        self.tree_model.setHorizontalHeaderLabels(["Raw file name", "N images"])
        self.treeView.setColumnWidth(0, 400)

        for i, file in enumerate(files):
            self.tree_model.appendRow(QtGui.QStandardItem(f"{file}"))
            self.tree_model.setItem(i, 1, QtGui.QStandardItem(f"{images[i]}"))

    def set_cal_file(self, file_path):
        if file_path is None:
            file_path = "undefined"
        self.cal_file_lbl.setText(file_path)
        self.cal_file_lbl.setToolTip("Calibration used for integration")

    def set_mask_file(self, file_path):
        if file_path is None:
            file_path = "undefined"
        self.mask_file_lbl.setText(file_path)
        self.mask_file_lbl.setToolTip("Mask used for integration")


class BatchFileControlWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchFileControlWidget, self).__init__()
        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)

        self.load_btn = FlatButton()
        self.load_btn.setToolTip("Load raw/proc data")
        self.load_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "open.ico")))
        self.load_btn.setIconSize(QtCore.QSize(13, 13))
        self.load_btn.setMaximumWidth(25)

        self.load_previous_folder_btn = FlatButton("<")
        self.load_previous_folder_btn.setToolTip("Load files in previous folder")

        self.load_next_folder_btn = FlatButton(">")
        self.load_next_folder_btn.setToolTip("Loads files in next folder")

        self.save_btn = FlatButton()
        self.save_btn.setToolTip("Save data")
        self.save_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "save.ico")))
        self.save_btn.setIconSize(QtCore.QSize(13, 13))
        self.save_btn.setMaximumWidth(25)

        self.folder_lbl = LabelExpandable("...")

        self.create_layout()

    def create_layout(self):
        self._layout.addWidget(self.load_btn)
        self._layout.addWidget(self.load_previous_folder_btn)
        self._layout.addWidget(self.load_next_folder_btn)
        self._layout.addWidget(self.save_btn)
        self._layout.addWidget(self.folder_lbl)

        self.setLayout(self._layout)


class BatchModeWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchModeWidget, self).__init__()
        self.setObjectName("batch_mode_widget")
        self._layout = QtWidgets.QHBoxLayout()

        self.view_f_btn = CheckableFlatButton("Files")
        self.view_2d_btn = CheckableFlatButton("2D")
        self.view_3d_btn = CheckableFlatButton("3D")

        self.unit_view_group = QtWidgets.QButtonGroup()
        self.unit_view_group.addButton(self.view_2d_btn)
        self.unit_view_group.addButton(self.view_3d_btn)
        self.unit_view_group.addButton(self.view_f_btn)

        self._layout.addWidget(self.view_f_btn)
        self._layout.addWidget(self.view_2d_btn)
        if open_gl:  # hide 3d plot when now opengl is present
            self._layout.addWidget(self.view_3d_btn)

        self.setLayout(self._layout)
        self.style_widgets()

    def style_widgets(self):
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

        mode_btn_height = 35
        mode_btn_width = 65

        mode_btns = [self.view_f_btn, self.view_2d_btn, self.view_3d_btn]

        for btn in mode_btns:
            btn.setWidth(mode_btn_width)
            btn.setHeight(mode_btn_height)


class BatchStackWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchStackWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()

        self.img_pg_layout = GraphicsLayoutWidget()
        self.img_view = IntegrationBatchWidget(
            self.img_pg_layout, orientation="horizontal"
        )

        self._layout.addWidget(self.img_pg_layout)
        self.setLayout(self._layout)
        self._layout.setContentsMargins(0, 0, 0, 0)


class BatchOptionsWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchOptionsWidget, self).__init__()
        self._layout = QtWidgets.QVBoxLayout()

        self.tth_btn = CheckableFlatButton("2θ")
        self.tth_btn.setChecked(True)
        self.q_btn = CheckableFlatButton("Q")
        self.d_btn = CheckableFlatButton("d")
        self.unit_btn_group = QtWidgets.QButtonGroup()
        self.unit_btn_group.addButton(self.tth_btn)
        self.unit_btn_group.addButton(self.q_btn)
        self.unit_btn_group.addButton(self.d_btn)
        self.background_btn = CheckableFlatButton("bg")
        self.bkg_cut_btn = CheckableFlatButton("T")
        self.bkg_cut_btn.setToolTip(
            "Trim data to show only region where background is calculated"
        )

        self.scale_log_btn = CheckableFlatButton("log")
        self.scale_sqrt_btn = CheckableFlatButton("√")
        self.scale_lin_btn = CheckableFlatButton("lin")
        self.scale_lin_btn.setChecked(True)
        self.unit_scale_group = QtWidgets.QButtonGroup()
        self.unit_scale_group.addButton(self.scale_log_btn)
        self.unit_scale_group.addButton(self.scale_sqrt_btn)
        self.unit_scale_group.addButton(self.scale_lin_btn)

        self.create_layout()
        self.style_widgets()
        self.set_tooltips()

    def create_layout(self):
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.tth_btn)
        self._layout.addWidget(self.q_btn)
        self._layout.addWidget(self.d_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.scale_lin_btn)
        self._layout.addWidget(self.scale_sqrt_btn)
        self._layout.addWidget(self.scale_log_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.background_btn)
        self._layout.addWidget(self.bkg_cut_btn)

        self.setLayout(self._layout)

    def style_widgets(self):
        self._layout.setContentsMargins(4, 0, 4, 6)
        self._layout.setSpacing(4)

        self.setStyleSheet(
            """QPushButton{
                padding: 0px;
                padding-right: 1px;
                border-radius: 3px;
            }
            """
        )

    def set_tooltips(self):
        self.scale_log_btn.setToolTip("Change scale to log")
        self.scale_sqrt_btn.setToolTip("Change scale to sqrt")
        self.scale_lin_btn.setToolTip("Change scale to linear")


class BatchSurfaceWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchSurfaceWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self.control_widget = BatchSurfaceViewNavigationWidget()

        self.surface_view = SurfaceWidget()
        self.pg_layout = self.surface_view.pg_layout

        self._layout.addWidget(self.control_widget)
        self._layout.addWidget(self.surface_view)
        self.setLayout(self._layout)

        self._layout.setContentsMargins(0, 0, 0, 0)


class BatchSurfaceViewNavigationWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchSurfaceViewNavigationWidget, self).__init__()

        self._layout = QtWidgets.QVBoxLayout()

        self.view3d_f_btn = FlatButton("F")
        self.view3d_s_btn = FlatButton("R")
        self.view3d_t_btn = FlatButton("T")
        self.view3d_i_btn = FlatButton("I")

        self.scale_x_btn = CheckableFlatButton("x")
        self.scale_y_btn = CheckableFlatButton("y")
        self.scale_z_btn = CheckableFlatButton("z")
        self.scale_s_btn = CheckableFlatButton("s")
        self.trim_h_btn = CheckableFlatButton("h")
        self.trim_l_btn = CheckableFlatButton("l")
        self.move_g_btn = CheckableFlatButton("g")
        self.move_m_btn = CheckableFlatButton("m")
        self.scale_s_btn.setChecked(True)

        self.scroll_btn_group = QtWidgets.QButtonGroup()
        self.scroll_btn_group.addButton(self.scale_x_btn)
        self.scroll_btn_group.addButton(self.scale_y_btn)
        self.scroll_btn_group.addButton(self.scale_z_btn)
        self.scroll_btn_group.addButton(self.scale_s_btn)
        self.scroll_btn_group.addButton(self.trim_h_btn)
        self.scroll_btn_group.addButton(self.trim_l_btn)
        self.scroll_btn_group.addButton(self.move_g_btn)
        self.scroll_btn_group.addButton(self.move_m_btn)

        self.m_color_btn = ColorButton()

        self.set_layout()
        self.style_widgets()
        self.set_tooltips()

    def set_layout(self):
        self._layout.addWidget(self.m_color_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.view3d_f_btn)
        self._layout.addWidget(self.view3d_s_btn)
        self._layout.addWidget(self.view3d_t_btn)
        self._layout.addWidget(self.view3d_i_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.scale_x_btn)
        self._layout.addWidget(self.scale_y_btn)
        self._layout.addWidget(self.scale_z_btn)
        self._layout.addWidget(self.scale_s_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.trim_h_btn)
        self._layout.addWidget(self.trim_l_btn)
        self._layout.addSpacerItem(VerticalSpacerItem())
        self._layout.addWidget(self.move_g_btn)
        self._layout.addWidget(self.move_m_btn)
        self.setLayout(self._layout)

    def set_tooltips(self):
        self.view3d_f_btn.setToolTip("Set front view")
        self.view3d_s_btn.setToolTip("Set right view")
        self.view3d_t_btn.setToolTip("Set top view")
        self.view3d_i_btn.setToolTip("Set isometric view")

        self.scale_x_btn.setToolTip("Scale plot in X direction")
        self.scale_y_btn.setToolTip("Scale plot in Y direction")
        self.scale_z_btn.setToolTip("Scale plot in Z direction")
        self.scale_s_btn.setToolTip("Scale plot")

        self.trim_h_btn.setToolTip("Cut higher values")
        self.trim_l_btn.setToolTip("Cut lower values")

        self.move_g_btn.setToolTip("Move grid along y axis")
        self.move_m_btn.setToolTip("Move marker along x axis")

        self.m_color_btn.setToolTip("Set marker color")

    def style_widgets(self):
        self._layout.setContentsMargins(4, 6, 4, 6)
        self._layout.setSpacing(4)

        self.setObjectName("pattern_left_control_widget")
        self.setMaximumWidth(30)
        self.setMinimumWidth(30)

        self.m_color_btn.setColor(0)
        self.m_color_btn.setMinimumHeight(80)

        self.setStyleSheet(
            """
                QPushButton{
                    padding: 0px;
                    padding-left: 1px;
                    border-radius: 3px;
                }
            """
        )


class BatchControlWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchControlWidget, self).__init__()
        self.integrate_btn = FlatButton("Integrate")
        self.load_proc_btn = FlatButton("Load proc data")

        self.waterfall_btn = CheckableFlatButton("Waterfall")

        self.calc_bkg_btn = FlatButton("Calc bkg")

        self.phases_btn = CheckableFlatButton("Show Phases")
        self.autoscale_btn = FlatButton("AutoScale")
        self.normalize_btn = FlatButton("Normalize")

        self._layout = QtWidgets.QHBoxLayout()

        self.create_layout()
        self.set_tooltips()
        self.style_widgets()

    def create_layout(self):
        self._layout.addWidget(self.integrate_btn)
        self._layout.addWidget(self.calc_bkg_btn)
        self._layout.addWidget(self.waterfall_btn)
        self._layout.addWidget(self.phases_btn)
        self._layout.addWidget(self.autoscale_btn)
        self._layout.addWidget(self.normalize_btn)

        self._layout.addSpacerItem(HorizontalSpacerItem())

        self.setLayout(self._layout)

    def set_tooltips(self):
        self.waterfall_btn.setToolTip("Create waterfall plot")
        self.calc_bkg_btn.setToolTip("Extract background")

    def style_widgets(self):
        self._layout.setContentsMargins(6, 6, 6, 6)
        self._layout.setSpacing(4)


class BatchImgPositionWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BatchImgPositionWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()

        self.step_series_widget = StepBatchWidget()
        self.step_raw_widget = StepBatchWidget()
        self.step_raw_widget.hide()
        self.mouse_pos_widget = MouseCurrentAndClickedWidget(CLICKED_COLOR)

        self._layout = QtWidgets.QHBoxLayout()
        self.create_layout()

    def create_layout(self):
        self._layout.setContentsMargins(4, 4, 4, 4)
        self._layout.addWidget(self.step_series_widget)
        self._layout.addWidget(self.step_raw_widget)
        self._layout.addSpacerItem(HorizontalSpacerItem())
        self._layout_mouse_pos = QtWidgets.QVBoxLayout()
        self._layout_mouse_pos.setContentsMargins(0, 0, 0, 0)
        self._layout_mouse_pos.addWidget(self.mouse_pos_widget)
        self._layout_mouse_pos.addStretch()
        self._layout.addLayout(self._layout_mouse_pos)

        self.setLayout(self._layout)


class StepBatchWidget(QtWidgets.QWidget):
    """
    Widget to navigate across frame in the batch

    Widget contains:
        buttons: next, previous
        slider
        fields: step, min, max, current
    """

    switch_frame = QtCore.Signal(int)

    iteration_name = ""

    def __init__(self):
        super(StepBatchWidget, self).__init__()
        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

        self.small_btn_max_width = 50
        self.small_btn_min_width = 20

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(5, 0, 0, 5)
        self.init_navigator()

        self.setLayout(self._layout)

    def init_navigator(self):
        self.next_btn = FlatButton(">")
        self.next_btn.setToolTip("Loads next {}".format(self.iteration_name))
        self.previous_btn = FlatButton("<")
        self.previous_btn.setToolTip(("Loads previous {}".format(self.iteration_name)))

        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)

        self.step_txt = QtWidgets.QSpinBox()
        self.step_txt.setValue(1)
        self.step_txt.setRange(1, 10000)

        self.stop_txt = QtWidgets.QSpinBox()
        self.stop_txt.setValue(0)
        self.stop_txt.setRange(0, 99000)

        self.start_txt = QtWidgets.QSpinBox()
        self.start_txt.setValue(0)
        self.start_txt.setRange(0, 99000)

        self.next_btn.setMaximumWidth(self.small_btn_max_width)
        self.previous_btn.setMaximumWidth(self.small_btn_max_width)
        self.next_btn.setMinimumWidth(self.small_btn_min_width)
        self.previous_btn.setMinimumWidth(self.small_btn_min_width)
        self.step_txt.setMinimumWidth(25)

        self._navigator_layout = QtWidgets.QGridLayout()
        self._navigator_layout.addWidget(self.previous_btn, 0, 0)
        self._navigator_layout.addWidget(self.slider, 0, 1)
        self._navigator_layout.addWidget(self.next_btn, 0, 2)

        self._step_layout = QtWidgets.QHBoxLayout()
        self._step_layout.addWidget(LabelAlignRight("Start:"))
        self._step_layout.addWidget(self.start_txt)
        self._step_layout.addWidget(LabelAlignRight("Stop:"))
        self._step_layout.addWidget(self.stop_txt)
        self._step_layout.addWidget(LabelAlignRight("Step:"))
        self._step_layout.addWidget(self.step_txt)
        self._navigator_layout.addLayout(self._step_layout, 1, 0, 1, 3)
        self._layout.addLayout(self._navigator_layout)

        self.pos_txt = QtWidgets.QLineEdit()
        self.pos_validator = QtGui.QIntValidator(1, 1)
        self.pos_txt.setText("0")
        self.pos_txt.setValidator(self.pos_validator)
        self.pos_txt.setToolTip("Currently loaded frame")
        self.pos_label = QtWidgets.QLabel("Frame:")
        self.pos_label.setToolTip("Number of frames: integrated/raw")
        self.pos_label.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignBottom)

        self._pos_layout = QtWidgets.QVBoxLayout()
        self._pos_layout.addWidget(self.pos_label)
        self._pos_layout.addWidget(self.pos_txt)
        self._layout.addLayout(self._pos_layout)

        self.pos_txt.setSizePolicy(
            QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Maximum
        )

        self.next_btn.clicked.connect(self.process_next_img)
        self.previous_btn.clicked.connect(self.process_prev_img)
        self.pos_txt.editingFinished.connect(self.process_pos_img)
        self.slider.sliderReleased.connect(self.process_slider)

    def process_next_img(self):
        step = self.step_txt.value()
        stop = self.stop_txt.value()
        pos = int(self.pos_txt.text())
        y = pos + step
        if y > stop:
            return
        self.pos_txt.setText(str(y))
        self.slider.setValue(y)
        self.switch_frame.emit(y)

    def process_prev_img(self):
        step = self.step_txt.value()
        start = self.start_txt.value()
        pos = int(self.pos_txt.text())
        y = pos - step
        if y < start:
            return
        self.pos_txt.setText(str(y))
        self.slider.setValue(y)
        self.switch_frame.emit(y)

    def process_pos_img(self):
        y = int(self.pos_txt.text())
        self.pos_txt.setText(str(y))
        self.slider.setValue(y)
        self.switch_frame.emit(y)

    def process_slider(self):
        y = self.slider.value()
        self.pos_txt.setText(str(y))
        self.switch_frame.emit(y)

    def get_image_range(self):
        step = self.step_txt.value()
        stop = self.stop_txt.value()
        start = self.start_txt.value()
        return start, stop, step



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import time
import os
from qtpy import QtWidgets

from ...model.util.ImgCorrection import (
    CbnCorrection,
    ObliqueAngleDetectorAbsorptionCorrection,
)

# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...widgets.UtilityWidgets import open_file_dialog
from ...model.DioptasModel import DioptasModel


class CorrectionController(object):
    """
    The CorrectionController manages the Correction controls in the integration window.
    """

    def __init__(self, widget, dioptas_model):
        """
        :param widget: Reference to IntegrationWidget
        :param dioptas_model: Reference to DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """

        self.widget = widget
        self.model = dioptas_model

        self.create_signals()

    def create_signals(self):
        # cbn correction
        self.widget.cbn_groupbox.clicked.connect(self.cbn_groupbox_changed)
        for row_ind in range(self.widget.cbn_param_tw.rowCount()):
            self.widget.cbn_param_tw.cellWidget(row_ind, 1).editingFinished.connect(
                self.cbn_groupbox_changed
            )
        self.widget.cbn_plot_btn.clicked.connect(self.cbn_plot_correction_btn_clicked)

        # oiadac correction
        self.widget.oiadac_groupbox.clicked.connect(self.oiadac_groupbox_changed)
        for row_ind in range(self.widget.oiadac_param_tw.rowCount()):
            self.widget.oiadac_param_tw.cellWidget(row_ind, 1).editingFinished.connect(
                self.oiadac_groupbox_changed
            )
        self.widget.oiadac_plot_btn.clicked.connect(self.oiadac_plot_btn_clicked)

        # transfer correction
        self.widget.transfer_load_original_btn.clicked.connect(
            self.transfer_load_original_btn_clicked
        )
        self.widget.transfer_load_response_btn.clicked.connect(
            self.transfer_load_response_btn_clicked
        )
        self.widget.transfer_plot_btn.clicked.connect(self.transfer_plot_btn_clicked)
        self.widget.transfer_gb.toggled.connect(self.transfer_gb_toggled)

        # general
        self.model.img_model.corrections_removed.connect(self.corrections_removed)

        # resetting plot buttons
        self.model.img_changed.connect(self.reset_plot_btns)
        self.model.cake_changed.connect(self.reset_plot_btns)

        # configurations
        self.model.configuration_selected.connect(self.update_gui)

    def transfer_load_original_btn_clicked(self):
        filename = open_file_dialog(
            self.widget,
            caption="Load Original Image File",
            directory=self.model.working_directories["image"],
        )
        if filename != "":
            self.widget.transfer_original_filename_lbl.setText(
                os.path.basename(filename)
            )
            self.model.img_model.transfer_correction.load_original_image(filename)
            self.model.img_model.enable_transfer_function()

    def transfer_load_response_btn_clicked(self):
        filename = open_file_dialog(
            self.widget,
            caption="Load Response Image File",
            directory=self.model.working_directories["image"],
        )
        if filename != "":
            self.widget.transfer_response_filename_lbl.setText(
                os.path.basename(filename)
            )
            self.model.img_model.transfer_correction.load_response_image(filename)
            self.model.img_model.enable_transfer_function()

    def transfer_plot_btn_clicked(self):
        if self.widget.transfer_plot_btn.isChecked():
            transfer_data = self.model.img_model.transfer_correction.get_data()
            if transfer_data is not None:
                self.widget.img_widget.plot_image(transfer_data, auto_level=True)
                self.widget.transfer_plot_btn.setText("Back")
            else:
                self.widget.transfer_plot_btn.setChecked(False)
        else:
            self.widget.transfer_plot_btn.setText("Plot")
            self.reset_img_widget()

    def update_transfer_widgets(self):
        original_filename = self.model.img_model.transfer_correction.original_filename
        response_filename = self.model.img_model.transfer_correction.response_filename
        if original_filename is not None:
            self.widget.transfer_original_filename_lbl.setText(
                os.path.basename(original_filename)
            )
        else:
            self.widget.transfer_original_filename_lbl.setText("None")
        if original_filename is not None:
            self.widget.transfer_response_filename_lbl.setText(
                os.path.basename(response_filename)
            )
        else:
            self.widget.transfer_response_filename_lbl.setText("None")

    def transfer_gb_toggled(self):
        if self.widget.transfer_gb.isChecked():
            self.model.img_model.enable_transfer_function()
        else:
            self.model.img_model.disable_transfer_function()

    def corrections_removed(self):
        self.widget.cbn_groupbox.setChecked(False)
        self.widget.oiadac_groupbox.setChecked(False)
        self.widget.transfer_gb.setChecked(False)
        self.widget.transfer_original_filename_lbl.setText("None")
        self.widget.transfer_response_filename_lbl.setText("None")
        QtWidgets.QMessageBox.critical(
            self.widget,
            "Shape Mismatch",
            "The loaded image and corrections have different shapes. "
            + "The corrections have been reset.",
        )

    def cbn_groupbox_changed(self):
        if not self.model.calibration_model.is_calibrated:
            self.widget.cbn_groupbox.setChecked(False)
            QtWidgets.QMessageBox.critical(
                self.widget,
                "ERROR",
                "Please calibrate the geometry first or load an existent calibration file. "
                + "The cBN seat correction needs a calibrated geometry.",
            )
            return

        if self.widget.cbn_groupbox.isChecked():
            diamond_thickness = self.widget.cbn_param_tw.cellWidget(0, 1).value()
            seat_thickness = self.widget.cbn_param_tw.cellWidget(1, 1).value()
            inner_seat_radius = self.widget.cbn_param_tw.cellWidget(2, 1).value()
            outer_seat_radius = self.widget.cbn_param_tw.cellWidget(3, 1).value()
            tilt = self.widget.cbn_param_tw.cellWidget(4, 1).value()
            tilt_rotation = self.widget.cbn_param_tw.cellWidget(5, 1).value()
            center_offset = self.widget.cbn_param_tw.cellWidget(6, 1).value()
            center_offset_angle = self.widget.cbn_param_tw.cellWidget(7, 1).value()
            seat_absorption_length = self.widget.cbn_param_tw.cellWidget(8, 1).value()
            anvil_absorption_length = self.widget.cbn_param_tw.cellWidget(9, 1).value()

            tth_array = (
                180.0 / np.pi * self.model.calibration_model.pattern_geometry.ttha
            )
            azi_array = (
                180.0 / np.pi * self.model.calibration_model.pattern_geometry.chia
            )

            new_cbn_correction = CbnCorrection(
                tth_array=tth_array,
                azi_array=azi_array,
                diamond_thickness=diamond_thickness,
                seat_thickness=seat_thickness,
                small_cbn_seat_radius=inner_seat_radius,
                large_cbn_seat_radius=outer_seat_radius,
                tilt=tilt,
                tilt_rotation=tilt_rotation,
                center_offset=center_offset,
                center_offset_angle=center_offset_angle,
                cbn_abs_length=seat_absorption_length,
                diamond_abs_length=anvil_absorption_length,
            )
            if not new_cbn_correction == self.model.img_model.get_img_correction("cbn"):
                t1 = time.time()
                new_cbn_correction.update()
                print(
                    "Time needed for correction calculation: {0}".format(
                        time.time() - t1
                    )
                )
                try:
                    self.model.img_model.delete_img_correction("cbn")
                except KeyError:
                    pass
                self.model.img_model.add_img_correction(new_cbn_correction, "cbn")
        else:
            self.model.img_model.delete_img_correction("cbn")

    def cbn_plot_correction_btn_clicked(self):
        if str(self.widget.cbn_plot_btn.text()) == "Plot":
            self.widget.img_widget.plot_image(
                self.model.img_model.img_corrections.get_correction("cbn").get_data(),
                True,
            )
            self.widget.cbn_plot_btn.setText("Back")
            self.widget.oiadac_plot_btn.setText("Plot")
        else:
            self.widget.cbn_plot_btn.setText("Plot")
            self.reset_img_widget()

    def update_cbn_widgets(self):
        params = self.model.img_model.img_corrections.get_correction("cbn").get_params()
        self.widget.cbn_param_tw.cellWidget(0, 1).setText(
            str(params["diamond_thickness"])
        )
        self.widget.cbn_param_tw.cellWidget(1, 1).setText(str(params["seat_thickness"]))
        self.widget.cbn_param_tw.cellWidget(2, 1).setText(
            str(params["small_cbn_seat_radius"])
        )
        self.widget.cbn_param_tw.cellWidget(3, 1).setText(
            str(params["large_cbn_seat_radius"])
        )
        self.widget.cbn_param_tw.cellWidget(4, 1).setText(str(params["tilt"]))
        self.widget.cbn_param_tw.cellWidget(5, 1).setText(str(params["tilt_rotation"]))
        self.widget.cbn_param_tw.cellWidget(6, 1).setText(
            str(params["diamond_abs_length"])
        )
        self.widget.cbn_param_tw.cellWidget(7, 1).setText(
            str(params["seat_abs_length"])
        )
        self.widget.cbn_param_tw.cellWidget(8, 1).setText(str(params["center_offset"]))
        self.widget.cbn_param_tw.cellWidget(9, 1).setText(
            str(params["center_offset_angle"])
        )
        self.widget.cbn_groupbox.setChecked(True)

    def oiadac_groupbox_changed(self):
        if not self.model.calibration_model.is_calibrated:
            self.widget.oiadac_groupbox.setChecked(False)
            QtWidgets.QMessageBox.critical(
                self.widget,
                "ERROR",
                "Please calibrate the geometry first or load an existent calibration file. "
                + "The oblique incidence angle detector absorption correction needs a calibrated"
                + "geometry.",
            )
            return

        if self.widget.oiadac_groupbox.isChecked():
            detector_thickness = self.widget.oiadac_param_tw.cellWidget(0, 1).value()
            absorption_length = self.widget.oiadac_param_tw.cellWidget(1, 1).value()

            _, fit2d_parameter = (
                self.model.calibration_model.get_calibration_parameter()
            )
            detector_tilt = fit2d_parameter["tilt"]
            detector_tilt_rotation = fit2d_parameter["tiltPlanRotation"]

            tth_array = self.model.calibration_model.pattern_geometry.ttha
            azi_array = self.model.calibration_model.pattern_geometry.chia
            import time

            t1 = time.time()

            oiadac_correction = ObliqueAngleDetectorAbsorptionCorrection(
                tth_array,
                azi_array,
                detector_thickness=detector_thickness,
                absorption_length=absorption_length,
                tilt=detector_tilt,
                rotation=detector_tilt_rotation,
            )
            print(
                "Time needed for correction calculation: {0}".format(time.time() - t1)
            )
            try:
                self.model.img_model.delete_img_correction("oiadac")
            except KeyError:
                pass
            self.model.img_model.add_img_correction(oiadac_correction, "oiadac")
        else:
            self.model.img_model.delete_img_correction("oiadac")

    def oiadac_plot_btn_clicked(self):
        if str(self.widget.oiadac_plot_btn.text()) == "Plot":
            self.widget.img_widget.plot_image(
                self.model.img_model._img_corrections.get_correction(
                    "oiadac"
                ).get_data(),
                True,
            )
            self.widget.oiadac_plot_btn.setText("Back")
            self.widget.cbn_plot_btn.setText("Plot")
        else:
            self.widget.oiadac_plot_btn.setText("Plot")
            self.reset_img_widget()

    def reset_img_widget(self):
        if self.widget.img_mode == "Cake":
            self.model.cake_changed.emit()
        elif self.widget.img_mode == "Image":
            self.model.img_changed.emit()

    def update_oiadac_widgets(self):
        params = self.model.img_model.img_corrections.get_correction(
            "oiadac"
        ).get_params()
        self.widget.oiadac_param_tw.cellWidget(0, 1).setText(
            str(params["detector_thickness"])
        )
        self.widget.oiadac_param_tw.cellWidget(1, 1).setText(
            str(params["absorption_length"])
        )
        self.widget.oiadac_groupbox.setChecked(True)

    def reset_plot_btns(self):
        self.widget.oiadac_plot_btn.setText("Plot")
        self.widget.oiadac_plot_btn.setChecked(False)
        self.widget.cbn_plot_btn.setText("Plot")
        self.widget.cbn_plot_btn.setChecked(False)
        self.widget.transfer_plot_btn.setText("Plot")
        self.widget.transfer_plot_btn.setChecked(False)

    def update_gui(self):
        if self.model.img_model.get_img_correction("cbn") is not None:
            self.update_cbn_widgets()
            self.widget.cbn_groupbox.blockSignals(True)
            self.widget.cbn_groupbox.setChecked(True)
            self.widget.cbn_groupbox.blockSignals(False)
        else:
            self.widget.cbn_groupbox.setChecked(False)

        if self.model.img_model.get_img_correction("oiadac") is not None:
            self.update_oiadac_widgets()
            self.widget.oiadac_groupbox.blockSignals(True)
            self.widget.oiadac_groupbox.setChecked(True)
            self.widget.oiadac_groupbox.blockSignals(False)
        else:
            self.widget.oiadac_groupbox.setChecked(False)

        if self.model.img_model.get_img_correction("transfer") is not None:
            self.update_transfer_widgets()
            # self.widget.transfer_gb.blockSignals(True)
            self.widget.transfer_gb.setChecked(True)
            # self.widget.transfer_gb.blockSignals(False)
        else:
            self.widget.transfer_gb.setChecked(False)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import contextmanager
from typing import Optional

import numpy as np
import pyqtgraph as pg


class Normalization:
    """Class defining the interface for implementing normalizations"""

    ID = ""
    """Abbreviation name of the normalization, must be unique among Normalization"""

    SHORTNAME = ""
    """Short name or symbol to use as e.g., button text"""

    DESCRIPTION = ""
    """Human-readable name of the normalization"""

    @staticmethod
    def apply(data: np.ndarray) -> np.ndarray:
        """Forward conversion of data to normalized data"""
        raise NotImplementedError()

    @staticmethod
    def revert(data: np.ndarray) -> np.ndarray:
        """Backward conversion of normalized data to data"""
        raise NotImplementedError()

    @staticmethod
    def invalid(data: np.ndarray) -> np.ndarray:
        """Element-wise invalidity test

        Returns a boolean array which is True for data elements that cannot be normalized.
        """
        return np.zeros(data.shape, dtype=bool)


class LinearNormalization(Normalization):
    ID = "linear"
    SHORTNAME = "lin"
    DESCRIPTION = "linear"

    @staticmethod
    def apply(data: np.ndarray) -> np.ndarray:
        return data

    @staticmethod
    def revert(data: np.ndarray) -> np.ndarray:
        return data


class LogNormalization(Normalization):
    ID = "log"
    SHORTNAME = "log"
    DESCRIPTION = "logarithmic"

    @staticmethod
    def apply(data: np.ndarray) -> np.ndarray:
        return np.log10(data)

    @staticmethod
    def revert(data: np.ndarray) -> np.ndarray:
        return np.power(10, data)

    @staticmethod
    def invalid(data: np.ndarray) -> np.ndarray:
        return data <= 0.0


class SqrtNormalization(Normalization):
    ID = "sqrt"
    SHORTNAME = "√"
    DESCRIPTION = "square root"

    @staticmethod
    def apply(data: np.ndarray) -> np.ndarray:
        return np.sqrt(data)

    @staticmethod
    def revert(data: np.ndarray) -> np.ndarray:
        return data * data

    @staticmethod
    def invalid(data: np.ndarray) -> np.ndarray:
        return data < 0.0


class NormalizedImageItem(pg.ImageItem):
    """pyqtgraph image item with support for data normalization"""

    _NORMALIZATIONS = dict((cls.ID, cls()) for cls in [
        LinearNormalization, LogNormalization, SqrtNormalization])
    """Dict of normalization name: Normalization instances"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__normalization = "linear"
        self.__rawImage = None

    @classmethod
    def supportedNormalizations(cls) -> tuple[str]:
        """Returns list of supported normalization names"""
        return tuple(cls._NORMALIZATIONS.keys())

    @classmethod
    def getNormalizationShortname(cls, normalization: str) -> str:
        """Returns a short string for given normalization name"""
        return cls._NORMALIZATIONS[normalization].SHORTNAME

    @classmethod
    def getNormalizationDescription(cls, normalization: str) -> str:
        """Returns a description string of given normalization name"""
        return cls._NORMALIZATIONS[normalization].DESCRIPTION

    def getData(self, copy: bool = True) -> Optional[np.ndarray]:
        """Returns the image data array

        :param copy: False to return the array used internally: do not modify!
        """
        if self.__rawImage is None:
            return None
        return np.array(self.__rawImage, copy=copy)

    def getNormalization(self) -> str:
        """Returns the currently used normalization"""
        return self.__normalization

    def setNormalization(self, normalization: str):
        """Set the data normalization to use"""
        if normalization not in self.supportedNormalizations():
            raise ValueError(f"Unsupported normalization: {normalization}")

        if normalization == self.__normalization:
            return

        # Get levels **before** changing the normalization
        levels = self.getLevels()
        self.__normalization = normalization
        self.setImage(image=self.__rawImage, levels=levels)

    def _getNorm(self) -> Normalization:
        return self._NORMALIZATIONS[self.getNormalization()]

    def setImage(self, image=None, *args, **kwargs):
        if image is None:
            return super().setImage(None, *args, **kwargs)

        self.__rawImage = image
        normalizedImage = self._getNorm().apply(image)
        return super().setImage(normalizedImage, *args, **kwargs)

    def getLevels(self):
        levels = super().getLevels()
        if levels is None:
            return None
        return self._getNorm().revert(levels)

    def setLevels(self, levels, update=True):
        if levels is None:
            return super().setLevels(levels, update)

        normalizedLevels = self._getNorm().apply(levels)
        return super().setLevels(normalizedLevels, update)

    @contextmanager
    def _useAsImage(self, image: Optional[np.ndarray]):
        """Context to temporarily use provided image as image attribute

        Used when calling ImageItem methods that needs to access a different
        image than the transformed one.
        """
        previousImage = self.image
        self.image = image
        try:
            yield
        finally:
            self.image = previousImage

    def quickMinMax(self, *args, **kwargs):
        with self._useAsImage(self.__rawImage):
            return super().quickMinMax(*args, **kwargs)

    def getHistogram(self, *args, **kwargs):
        with self._useAsImage(self.__rawImage):
            return super().getHistogram(*args, **kwargs)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque

import fabio
import numpy as np
import skimage.draw
from PIL import Image
from qtpy import QtCore
from math import sqrt, atan2, cos, sin

from .util.cosmics import cosmicsimage


class MaskModel(object):
    def __init__(self, mask_dimension=(2048, 2048)):
        self.mask_dimension = mask_dimension
        self.reset_dimension()
        self.filename = ''
        self.mode = True
        self.roi = None

        self._mask_data = np.zeros(self.mask_dimension, dtype=bool)
        self._undo_deque = deque(maxlen=50)
        self._redo_deque = deque(maxlen=50)

    def set_dimension(self, mask_dimension):
        if not np.array_equal(mask_dimension, self.mask_dimension):
            self.mask_dimension = mask_dimension
            self.reset_dimension()

    def reset_dimension(self):
        if self.mask_dimension is not None:
            self._mask_data = np.zeros(self.mask_dimension, dtype=bool)
            self._undo_deque = deque(maxlen=50)
            self._redo_deque = deque(maxlen=50)

    @property
    def roi_mask(self):
        if self.roi is not None:
            roi_mask = np.ones(self.mask_dimension)
            x1, x2, y1, y2 = self.roi
            if x1 < 0:
                x1 = 0
            if y1 < 0:
                y1 = 0
            roi_mask[int(x1):int(x2), int(y1):int(y2)] = 0

            return roi_mask

        else:
            return None

    def get_mask(self):
        if self.roi is None:
            return self._mask_data
        elif self.roi is not None:
            return np.logical_or(self._mask_data, self.roi_mask)

    def get_img(self):
        return self._mask_data

    def update_deque(self):
        """
        Saves the current mask data into a deque, which can be popped later
        to provide an undo/redo feature.
        When performing a new action the old redo steps will be cleared..._
        """
        self._undo_deque.append(np.copy(self._mask_data))
        self._redo_deque.clear()

    def undo(self):
        try:
            old_data = self._undo_deque.pop()
            self._redo_deque.append(np.copy(self._mask_data))
            self._mask_data = old_data
        except IndexError:
            pass

    def redo(self):
        try:
            new_data = self._redo_deque.pop()
            self._undo_deque.append(np.copy(self._mask_data))
            self._mask_data = new_data
        except IndexError:
            pass

    def mask_below_threshold(self, img_data, threshold):
        self.update_deque()
        self._mask_data += (img_data < threshold)

    def mask_above_threshold(self, img_data, threshold):
        self.update_deque()
        self._mask_data += (img_data > threshold)

    def mask_QGraphicsRectItem(self, QGraphicsRectItem):
        rect = QGraphicsRectItem.rect()
        self.mask_rect(rect.top(), rect.left(), rect.height(), rect.width())

    def mask_QGraphicsPolygonItem(self, QGraphicsPolygonItem):
        """
        Masks a polygon given by a QGraphicsPolygonItem from the QtWidgets Library.
        Uses the sklimage.draw.polygon function.
        """

        # get polygon points
        poly_list = list(QGraphicsPolygonItem.vertices)
        x = np.zeros(len(poly_list))
        y = np.zeros(len(poly_list))

        for i, point in enumerate(poly_list):
            x[i] = point.x()
            y[i] = point.y()
        self.mask_polygon(x, y)

    def mask_QGraphicsEllipseItem(self, QGraphicsEllipseItem):
        """
        Masks an Ellipse given by a QGraphicsEllipseItem from the QtWidgets
        Library. Uses the skimage.draw.ellipse function.
        """
        bounding_rect = QGraphicsEllipseItem.rect()
        cx = bounding_rect.center().x()
        cy = bounding_rect.center().y()
        x_radius = bounding_rect.width() * 0.5
        y_radius = bounding_rect.height() * 0.5
        self.mask_ellipse(int(cx), int(cy), int(x_radius), int(y_radius))

    def mask_rect(self, x, y, width, height):
        """
        Masks a rectangle. x and y parameters are the upper left corner
        of the rectangle.
        """
        self.update_deque()
        if width > 0:
            x_ind1 = np.round(x)
            x_ind2 = np.round(x + width)
        else:
            x_ind1 = np.round(x + width)
            x_ind2 = np.round(x)
        if height > 0:
            y_ind1 = np.round(y)
            y_ind2 = np.round(y + height)
        else:
            y_ind1 = np.round(y + height)
            y_ind2 = np.round(y)

        if x_ind1 < 0:
            x_ind1 = 0
        if y_ind1 < 0:
            y_ind1 = 0

        x_ind1, x_ind2, y_ind1, y_ind2 = int(x_ind1), int(x_ind2), int(y_ind1), int(y_ind2)
        self._mask_data[x_ind1:x_ind2, y_ind1:y_ind2] = self.mode

    def mask_polygon(self, x, y):
        """
        Masks the a polygon with given vertices. x and y are lists of
        the polygon vertices. Uses the draw.polygon implementation of
        the skimage library.
        """
        self.update_deque()
        rr, cc = skimage.draw.polygon(y, x, self._mask_data.shape)
        self._mask_data[rr, cc] = self.mode

    def mask_ellipse(self, cx, cy, x_radius, y_radius):
        """
        Masks an ellipse with center coordinates (cx, cy) and the radii
        given. Uses the draw.ellipse implementation of
        the skimage library.
        """
        self.update_deque()
        rr, cc = skimage.draw.ellipse(
            cy, cx, y_radius, x_radius, shape=self._mask_data.shape)
        self._mask_data[rr, cc] = self.mode

    def grow(self):
        self.update_deque()
        self._mask_data[1:, :] = np.logical_or(self._mask_data[1:, :], self._mask_data[:-1, :])
        self._mask_data[:-1, :] = np.logical_or(self._mask_data[:-1, :], self._mask_data[1:, :])
        self._mask_data[:, 1:] = np.logical_or(self._mask_data[:, 1:], self._mask_data[:, :-1])
        self._mask_data[:, :-1] = np.logical_or(self._mask_data[:, :-1], self._mask_data[:, 1:])

    def shrink(self):
        self.update_deque()
        self._mask_data[1:, :] = np.logical_and(self._mask_data[1:, :], self._mask_data[:-1, :])
        self._mask_data[:-1, :] = np.logical_and(self._mask_data[:-1, :], self._mask_data[1:, :])
        self._mask_data[:, 1:] = np.logical_and(self._mask_data[:, 1:], self._mask_data[:, :-1])
        self._mask_data[:, :-1] = np.logical_and(self._mask_data[:, :-1], self._mask_data[:, 1:])

    def invert_mask(self):
        self.update_deque()
        self._mask_data = np.logical_not(self._mask_data)

    def clear_mask(self):
        self.update_deque()
        self._mask_data[:, :] = False

    def remove_cosmic(self, img):
        self.update_deque()
        test = cosmicsimage(img, sigclip=3.0, objlim=3.0)
        num = 2
        for i in range(num):
            test.lacosmiciteration(True)
            test.clean()
            self._mask_data = np.logical_or(self._mask_data, np.array(test.mask, dtype='bool'))

    def set_mode(self, mode):
        """
        sets the mode to unmask or mask which equals mode = False or True
        """
        self.mode = mode

    def set_mask(self, mask_data):
        self.update_deque()
        self._mask_data = mask_data

    def save_mask(self, filename: str, flipud: bool = False):
        """Save current mask to file

        :param filename: Path of the file to write
        :param flipud: True to apply a vertical flip before saving the mask
        """
        im_array = np.int8(self.get_img())
        if flipud:
            im_array = np.flipud(im_array)

        if filename.endswith('.npy'):
            np.save(filename, im_array)
        elif filename.endswith('.edf'):
            fabio.edfimage.EdfImage(im_array).write(filename)
        else:
            im = Image.fromarray(im_array)
            try:
                im.save(filename, "tiff", compression="tiff_deflate")
            except OSError:
                try:
                    im.save(filename, "tiff", compression="tiff_adobe_deflate")
                except IOError:
                    im.save(filename, "tiff")

        self.filename = filename

    @staticmethod
    def read_mask_file(filename: str, flipud: bool = False) -> np.ndarray:
        """Load an image mask from file.

        :param filename: Path to the file to read
        :param flipud: True to apply a vertical flip to the mask
        """
        if filename.endswith('.npy'):
            data = np.load(filename)
        elif filename.endswith('.edf'):
            data = fabio.open(filename).data
        else:
            try:
                data = np.array(Image.open(filename))
            except IOError:
                data = np.loadtxt(filename)

        if flipud:
            data = np.flipud(data)
        return data

    def load_mask(self, filename: str, flipud: bool = False):
        """Load mask from file and replace the current one

        :param filename: Path to the file to read
        :param flipud: True to apply a vertical flip to the loaded mask
        """
        data = self.read_mask_file(filename, flipud)

        if self.mask_dimension == data.shape:
            self.filename = filename
            self.mask_dimension = data.shape
            self.reset_dimension()
            self.set_mask(data)
            return True
        return False

    def add_mask(self, filename: str, flipud: bool = False):
        """Combine mask loaded from file with the current one

        :param filename: Path to the file to read
        :param flipud: True to apply a vertical flip to the loaded mask
        """
        data = self.read_mask_file(filename, flipud)

        if self.get_mask().shape == data.shape:
            self._add_mask(data)
            return True
        return False

    def _add_mask(self, mask_data):
        self.update_deque()
        self._mask_data = np.logical_or(self._mask_data,
                                        np.array(mask_data, dtype='bool'))

    def find_center_of_circle_from_three_points(self, a, b, c):
        xa, ya = a.x(), a.y()
        xb, yb = b.x(), b.y()
        xc, yc = c.x(), c.y()
        # if (xa == xb and ya == yb) or (xa == xc and ya == yc) or (xb == xc and yb == yc):
        #     return None
        mid_ab_x = (xa + xb) / 2.0
        mid_ab_y = (ya + yb) / 2.0
        mid_bc_x = (xb + xc) / 2.0
        mid_bc_y = (yb + yc) / 2.0
        slope_ab = (yb - ya) / (xb - xa)
        slope_bc = (yc - yb) / (xc - xb)
        slope_p_ab = -1.0 / slope_ab
        slope_p_bc = -1.0 / slope_bc
        b_p_ab = mid_ab_y - slope_p_ab * mid_ab_x
        b_p_bc = mid_bc_y - slope_p_bc * mid_bc_x
        x0 = (b_p_bc - b_p_ab) / (slope_p_ab - slope_p_bc)
        y0 = slope_p_ab * x0 + b_p_ab
        self.center_for_arc = QtCore.QPointF(x0, y0)
        return self.center_for_arc

    @staticmethod
    def find_radius_of_circle_from_center_and_point(p0, a):
        r = sqrt((a.x() - p0.x()) ** 2 + (a.y() - p0.y()) ** 2)
        return r

    def find_n_angles_on_arc_from_three_points_around_p0(self, p0, pa, pb, pc, n):
        phi_a = self.calc_angle_from_center_and_point(p0, pa)
        phi_b = self.calc_angle_from_center_and_point(p0, pb)
        phi_c = self.calc_angle_from_center_and_point(p0, pc)
        if phi_c < phi_a < phi_b or phi_b < phi_c < phi_a:
            phi_range = np.linspace(phi_a, phi_c + 2 * np.pi, n)
        elif phi_a < phi_b < phi_c or phi_c < phi_b < phi_a:
            phi_range = np.linspace(phi_a, phi_c, n)
        elif phi_a < phi_c < phi_b or phi_b < phi_a < phi_c:
            phi_range = np.linspace(phi_a + 2 * np.pi, phi_c, n)
        else:
            return None
        return phi_range

    @staticmethod
    def calc_angle_from_center_and_point(p0, pa):
        phi = atan2(pa.y() - p0.y(), pa.x() - p0.x())
        return phi

    @staticmethod
    def calc_arc_points_from_angles(p0, r, width, phi_range):
        p = []
        for phi in phi_range:
            xn = p0.x() + (r - width) * cos(phi)
            yn = p0.y() + (r - width) * sin(phi)
            p.append(QtCore.QPointF(xn, yn))
        return p



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dioptas.model.OverlayModel import OverlayModel
from dioptas.widgets.plot_widgets import PatternWidget


class OverlayInPatternController(object):
    def __init__(self, pattern_widget: PatternWidget, overlay_model: OverlayModel):
        self.model = overlay_model
        self.pattern_widget = pattern_widget

        self.connect()

    def connect(self):
        self.model.overlay_added.connect(self.overlay_added)
        self.model.overlay_removed.connect(self.overlay_removed)
        self.model.overlay_changed.connect(self.overlay_changed)

    def overlay_added(self):
        overlay = self.model.get_overlay(len(self.model.overlays) - 1)
        color = self.model.get_overlay_color(len(self.model.overlays) - 1)
        if overlay is not None:
            self.pattern_widget.add_overlay(overlay, color)

    def overlay_removed(self, index: int):
        self.pattern_widget.remove_overlay(index)

    def overlay_changed(self, index: int):
        overlay = self.model.get_overlay(index)
        if overlay is not None:
            self.pattern_widget.update_overlay(index, overlay)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from qtpy import QtWidgets, QtCore, QtGui

from ....model.PhaseModel import PhaseLoadError
from ....model.util.HelperModule import get_base_name
from ....controller.integration.phase.JcpdsEditorController import JcpdsEditorController
from ....widgets.UtilityWidgets import save_file_dialog, open_file_dialog, open_files_dialog

from .PhaseInPatternController import PhaseInPatternController
from .PhaseInCakeController import PhaseInCakeController
from .PhaseInBatchController import PhaseInBatchController

# imports for type hinting in PyCharm -- DO NOT DELETE
from ....model.DioptasModel import DioptasModel
from ....widgets.integration import IntegrationWidget
from ....widgets.UtilityWidgets import CifConversionParametersDialog


class PhaseController(object):
    """
    PhaseController handles all the interaction between the phase controls in the Integration View and the
    PhaseModel object.
    """

    def __init__(self, integration_widget, dioptas_model):
        """
        :param integration_widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object

        :type integration_widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """
        self.integration_widget = integration_widget
        self.phase_widget = self.integration_widget.phase_widget
        self.pattern_widget = self.integration_widget.pattern_widget
        self.model = dioptas_model

        self.cif_conversion_dialog = CifConversionParametersDialog(self.integration_widget)

        self.phase_in_pattern_controller = PhaseInPatternController(self.integration_widget.pattern_widget, dioptas_model)
        self.phase_in_cake_controller = PhaseInCakeController(self.integration_widget, dioptas_model)
        self.phase_in_batch_controller = PhaseInBatchController(self.integration_widget.batch_widget, dioptas_model)
        self.jcpds_editor_controller = JcpdsEditorController(self.integration_widget, self.model)

        self.phase_lw_items = []
        self.create_signals()
        self.update_temperature_step()
        self.update_pressure_step()

    def create_signals(self):
        # Button Callbacks
        self.phase_widget.add_btn.clicked.connect(self.add_btn_click_callback)
        self.phase_widget.delete_btn.clicked.connect(self.delete_btn_click_callback)
        self.phase_widget.clear_btn.clicked.connect(self.clear_phases)
        self.phase_widget.save_list_btn.clicked.connect(self.save_btn_clicked_callback)
        self.phase_widget.load_list_btn.clicked.connect(self.load_list_btn_clicked_callback)

        # Spinbox Callbacks
        self.phase_widget.pressure_step_msb.valueChanged.connect(self.update_pressure_step)
        self.phase_widget.temperature_step_msb.valueChanged.connect(self.update_temperature_step)

        self.phase_widget.pressure_sb_value_changed.connect(self.model.phase_model.set_pressure)
        self.phase_widget.temperature_sb_value_changed.connect(self.model.phase_model.set_temperature)

        # Color and State
        self.phase_widget.color_btn_clicked.connect(self.color_btn_clicked)
        self.phase_widget.show_cb_state_changed.connect(self.model.phase_model.set_phase_visible)
        self.phase_widget.apply_to_all_cb.stateChanged.connect(self.apply_to_all_callback)

        # TableWidget
        self.phase_widget.phase_tw.horizontalHeader().sectionClicked.connect(self.phase_tw_header_section_clicked)

        # PhaseModel Signals
        self.model.phase_model.phase_added.connect(self.phase_added)
        self.model.phase_model.phase_changed.connect(self.phase_changed)
        self.model.phase_model.phase_removed.connect(self.phase_removed)

    def connect_click_function(self, emitter, function):
        emitter.clicked.connect(function)

    def add_btn_click_callback(self, *args, **kwargs):
        """
        Loads a new phase from jcpds file.
        """
        if not self.model.calibration_model.is_calibrated:
            self.integration_widget.show_error_msg("Can not load phase without calibration.")

        filenames = [kwargs.get('filenames', None)]

        if filenames[0] is None:
            filenames = open_files_dialog(self.integration_widget, "Load Phase(s).",
                                          self.model.working_directories['phase'])

        if len(filenames):
            self.model.working_directories['phase'] = os.path.dirname(str(filenames[0]))
            progress_dialog = QtWidgets.QProgressDialog("Loading multiple phases.", "Abort Loading", 0, len(filenames),
                                                        self.integration_widget)
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
            progress_dialog.setWindowFlags(QtCore.Qt.FramelessWindowHint)
            progress_dialog.show()
            QtWidgets.QApplication.processEvents()
            for ind, filename in enumerate(filenames):
                filename = str(filename)
                progress_dialog.setValue(ind)
                progress_dialog.setLabelText("Loading: " + os.path.basename(filename))
                QtWidgets.QApplication.processEvents()

                self._add_phase(filename)

                if progress_dialog.wasCanceled():
                    break
            progress_dialog.close()
            QtWidgets.QApplication.processEvents()

    def _add_phase(self, filename):
        try:
            if filename.endswith("jcpds"):
                self.model.phase_model.add_jcpds(filename)
            elif filename.endswith(".cif"):
                self.cif_conversion_dialog.exec_()
                self.model.phase_model.add_cif(filename,
                                               self.cif_conversion_dialog.int_cutoff,
                                               self.cif_conversion_dialog.min_d_spacing)
        except PhaseLoadError as e:
            self.integration_widget.show_error_msg(
                'Could not load:\n\n{}.\n\nPlease check if the format of the input file is correct.'. \
                    format(e.filename))

    def phase_added(self):
        color = self.model.phase_model.phase_colors[-1]
        self.phase_widget.add_phase(get_base_name(self.model.phase_model.phase_files[-1]),
                                    '#%02x%02x%02x' % (int(color[0]), int(color[1]), int(color[2])))

    def phase_changed(self, ind):
        phase_name = get_base_name(self.model.phase_model.phases[ind].filename)
        if self.model.phase_model.phases[ind].params['modified']:
            phase_name += '*'
        self.phase_widget.rename_phase(ind, phase_name)
        self.phase_widget.set_phase_pressure(ind, self.model.phase_model.phases[ind].params['pressure'])
        self.phase_widget.set_phase_temperature(ind, self.model.phase_model.phases[ind].params['temperature'])
        self.phase_widget.temperature_sbs[ind].setEnabled(int(self.model.phase_model.phases[ind].has_thermal_expansion()))

    def delete_btn_click_callback(self):
        """
        Deletes the currently selected Phase
        """
        cur_ind = self.phase_widget.get_selected_phase_row()
        if cur_ind >= 0:
            self.model.phase_model.del_phase(cur_ind)

    def phase_removed(self, ind):
        self.phase_widget.del_phase(ind)
        # self.img_view_widget.del_cake_phase(ind)

        if self.jcpds_editor_controller.active:
            ind = self.phase_widget.get_selected_phase_row()
            if ind >= 0:
                self.jcpds_editor_controller.show_phase(self.model.phase_model.phases[ind])
            else:
                self.jcpds_editor_controller.jcpds_widget.close()

    def load_list_btn_clicked_callback(self):
        filename = open_file_dialog(self.integration_widget, caption="Load Phase List",
                                    directory=self.model.working_directories['phase'],
                                    filter="*.txt")
        if filename == '':
            return
        with open(filename, 'r') as phase_file:
            if phase_file == '':
                return
            for line in phase_file.readlines():
                line = line.replace('\n', '')
                phase, use_flag, color, name, pressure, temperature = line.split(',')
                self.add_btn_click_callback(filenames=phase)
                row = self.phase_widget.phase_tw.rowCount() - 1
                self.phase_widget.phase_show_cbs[row].setChecked(bool(use_flag))
                self.phase_widget.phase_color_btns[row].setStyleSheet(
                    f"background-color: {color}; margin: 2px;"
                )
                self.pattern_widget.set_phase_color(row, color)
                self.phase_widget.phase_tw.item(row, 2).setText(name)
                self.phase_widget.set_phase_pressure(row, float(pressure))
                self.model.phase_model.set_pressure(row, float(pressure))
                temperature = float(temperature)

                if temperature != '':
                    self.phase_widget.set_phase_temperature(row, temperature)
                    self.model.phase_model.set_temperature(row, temperature)

    def save_btn_clicked_callback(self):
        if len(self.model.phase_model.phase_files) < 1:
            return
        filename = save_file_dialog(self.integration_widget, "Save Phase List.",
                                    os.path.join(self.model.working_directories['phase'], 'phase_list.txt'),
                                    'Text (*.txt)')

        if filename == '':
            return

        with open(filename, 'w') as phase_file:
            for file_name, phase_cb, color_btn, row in zip(self.model.phase_model.phase_files,
                                                           self.phase_widget.phase_show_cbs,
                                                           self.phase_widget.phase_color_btns,
                                                           range(self.phase_widget.phase_tw.rowCount())):
                phase_file.write(file_name + ',' + str(phase_cb.isChecked()) + ',' +
                                 color_btn.styleSheet().split(";")[0].replace('background-color:', '').replace(' ', '') + ',' +
                                 self.phase_widget.phase_tw.item(row, 2).text() + ',' +
                                 self.phase_widget.pressure_sbs[row].text() + ',' +
                                 self.phase_widget.temperature_sbs[row].text() + '\n')

    def clear_phases(self):
        """
        Deletes all phases from the GUI and phase data
        """
        while self.phase_widget.phase_tw.rowCount() > 0:
            self.delete_btn_click_callback()
            self.jcpds_editor_controller.close_view()

    def update_pressure_step(self):
        for pressure_sb in self.phase_widget.pressure_sbs:
            pressure_sb.setSingleStep(self.phase_widget.pressure_step_msb.value())

    def update_temperature_step(self):
        for temperature_sb in self.phase_widget.temperature_sbs:
            temperature_sb.setSingleStep(self.phase_widget.temperature_step_msb.value())

    def color_btn_clicked(self, ind, button):
        previous_color = button.palette().color(QtGui.QPalette.Button)
        new_color = QtWidgets.QColorDialog.getColor(previous_color, self.integration_widget)
        if new_color.isValid():
            color = new_color.toRgb()
        else:
            color = previous_color.toRgb()
        self.model.phase_model.set_color(ind, (color.red(), color.green(), color.blue()))
        button.setStyleSheet(f"background-color: {color.name()}; margin: 2px;")

    def apply_to_all_callback(self):
        self.model.phase_model.same_conditions = self.phase_widget.apply_to_all_cb.isChecked()

    def phase_tw_header_section_clicked(self, ind):
        if ind != 0:
            return

        current_checkbox_state = False
        # check whether any checkbox is checked, if one is true current_checkbox_state will be True too
        for cb in self.phase_widget.phase_show_cbs:
            current_checkbox_state = current_checkbox_state or cb.isChecked()

        # assign the the opposite to all checkboxes
        for cb in self.phase_widget.phase_show_cbs:
            cb.setChecked(not current_checkbox_state)

    def get_unit(self):
        """
        returns the unit currently selected in the GUI
                possible values: 'tth', 'q', 'd'
        """
        if self.integration_widget.pattern_tth_btn.isChecked():
            return 'tth'
        elif self.integration_widget.pattern_q_btn.isChecked():
            return 'q'
        elif self.integration_widget.pattern_d_btn.isChecked():
            return 'd'



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from glob import glob
import os
import typing

from qtpy.QtGui import QCloseEvent, QColorSpace, QMouseEvent
import numpy as np
import h5py
from qtpy import QtWidgets, QtCore
from pyqtgraph import makeQImage

from ...model.util.HelperModule import get_partial_value, get_partial_index
from ...widgets.UtilityWidgets import (
    get_progress_dialog,
    open_files_dialog,
    save_file_dialog,
)
from ...widgets.integration.BatchWidget import open_gl

# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel


class BatchController(object):
    """
    The class manages the Image actions in the batch integration Window. It connects the file actions, as
    well as interaction with the image_view.
    """

    def __init__(self, widget: IntegrationWidget, dioptas_model: DioptasModel):
        """
        :param widget: Reference to IntegrationView
        :param dioptas_model: Reference to DioptasModel object
        """
        self.widget = widget
        self.model = dioptas_model

        self.clicks = 0
        self.rect = None
        self.scale = np.array

        self.create_signals()
        self.create_mouse_behavior()

        self.min_val = {"lin": 0, "sqrt": 0.1, "log": 0.1, "current": 0}
        self.size_threshold = 500000

    def create_signals(self):
        """
        Creates all the connections of the GUI elements.
        """
        self.widget.integration_control_widget.img_control_widget.batch_btn.clicked.connect(
            self.show_batch_frame
        )
        self.widget.batch_widget.closeEvent = self.close_batch_frame

        self.widget.batch_widget.file_control_widget.load_btn.clicked.connect(
            self.load_data
        )
        self.widget.batch_widget.file_control_widget.load_previous_folder_btn.clicked.connect(
            self.load_previous_folder
        )
        self.widget.batch_widget.file_control_widget.load_next_folder_btn.clicked.connect(
            self.load_next_folder
        )
        self.widget.batch_widget.file_control_widget.save_btn.clicked.connect(
            self.save_data
        )

        self.widget.batch_widget.control_widget.integrate_btn.clicked.connect(
            self.integrate
        )
        self.widget.batch_widget.control_widget.waterfall_btn.clicked.connect(
            self.waterfall_mode
        )
        self.widget.batch_widget.control_widget.phases_btn.clicked.connect(
            self.toggle_show_phases
        )
        self.widget.batch_widget.mode_widget.view_f_btn.clicked.connect(
            self.change_view
        )
        self.widget.batch_widget.mode_widget.view_2d_btn.clicked.connect(
            self.change_view
        )
        self.widget.batch_widget.options_widget.scale_log_btn.mouseReleaseEvent = (
            self.change_scale_log
        )
        self.widget.batch_widget.options_widget.scale_lin_btn.mouseReleaseEvent = (
            self.change_scale_lin
        )
        self.widget.batch_widget.options_widget.scale_sqrt_btn.mouseReleaseEvent = (
            self.change_scale_sqrt
        )
        self.widget.batch_widget.options_widget.background_btn.clicked.connect(
            self.subtract_background
        )
        self.widget.batch_widget.options_widget.bkg_cut_btn.clicked.connect(
            lambda: self.plot_batch()
        )
        self.widget.batch_widget.control_widget.calc_bkg_btn.clicked.connect(
            self.extract_background
        )
        self.widget.batch_widget.control_widget.autoscale_btn.clicked.connect(
            self.img_autoscale_btn_clicked
        )
        self.widget.batch_widget.control_widget.normalize_btn.clicked.connect(
            self.normalize_btn_clicked
        )

        # set unit of x axis
        self.widget.batch_widget.options_widget.tth_btn.clicked.connect(
            self.set_unit_tth
        )
        self.widget.batch_widget.options_widget.q_btn.clicked.connect(self.set_unit_q)
        self.widget.batch_widget.options_widget.d_btn.clicked.connect(self.set_unit_d)
        self.widget.pattern_q_btn.clicked.connect(self.set_unit_q)
        self.widget.pattern_tth_btn.clicked.connect(self.set_unit_tth)
        self.widget.pattern_d_btn.clicked.connect(self.set_unit_d)

        # work with filenames
        self.widget.img_directory_txt.editingFinished.connect(
            self.directory_txt_changed
        )
        self.widget.img_directory_btn.clicked.connect(self.directory_txt_changed)

        # image navigation
        self.widget.batch_widget.position_widget.step_raw_widget.switch_frame.connect(
            self.switch_frame
        )
        self.widget.batch_widget.position_widget.step_series_widget.switch_frame.connect(
            self.switch_frame
        )
        self.widget.batch_widget.position_widget.step_series_widget.start_txt.valueChanged.connect(
            self.set_range_img
        )
        self.widget.batch_widget.position_widget.step_series_widget.stop_txt.valueChanged.connect(
            self.set_range_img
        )
        self.widget.batch_widget.position_widget.step_series_widget.step_txt.valueChanged.connect(
            self.process_step
        )

        # Surface widget signals
        if open_gl:
            self.widget.batch_widget.mode_widget.view_3d_btn.clicked.connect(
                self.change_view
            )
            surface_navigation_widget = (
                self.widget.batch_widget.surface_widget.control_widget
            )
            surface_navigation_widget.view3d_f_btn.clicked.connect(self.set_3d_view_f)
            surface_navigation_widget.view3d_s_btn.clicked.connect(self.set_3d_view_s)
            surface_navigation_widget.view3d_t_btn.clicked.connect(self.set_3d_view_t)
            surface_navigation_widget.view3d_i_btn.clicked.connect(self.set_3d_view_i)
            surface_navigation_widget.scale_x_btn.clicked.connect(self.pressed_button_x)
            surface_navigation_widget.scale_y_btn.clicked.connect(self.pressed_button_y)
            surface_navigation_widget.scale_z_btn.clicked.connect(self.pressed_button_z)
            surface_navigation_widget.scale_s_btn.clicked.connect(self.pressed_button_s)
            surface_navigation_widget.trim_h_btn.clicked.connect(self.pressed_button_h)
            surface_navigation_widget.trim_l_btn.clicked.connect(self.pressed_button_l)
            surface_navigation_widget.move_g_btn.clicked.connect(self.pressed_button_g)
            surface_navigation_widget.move_m_btn.mouseReleaseEvent = (
                self.pressed_button_m
            )
            surface_navigation_widget.m_color_btn.sigColorChanged.connect(
                self.set_marker_color
            )

        self.widget.batch_widget.stack_plot_widget.img_view.img_view_box.sigRangeChanged.connect(
            self.update_axes_range
        )
        self.model.configuration_selected.connect(self.update_gui)

    def create_mouse_behavior(self):
        """
        Creates the signal connections of mouse interactions
        """
        self.widget.batch_widget.stack_plot_widget.img_view.mouse_moved.connect(
            self.show_img_mouse_position
        )
        self.widget.batch_widget.stack_plot_widget.img_view.mouse_left_clicked.connect(
            self.img_mouse_click
        )

        self.model.clicked_tth_changed.connect(self.update_vertical_line_pos)

        # 3D
        if open_gl:
            self.widget.batch_widget.surface_widget.pg_layout.wheelEvent = (
                self.wheel_event_3d
            )
            self.widget.batch_widget.surface_widget.pg_layout.keyPressEvent = (
                self.key_pressed_3d
            )

    def show_batch_frame(self):
        self.widget.batch_widget.raise_widget()

    def close_batch_frame(self, a0: typing.Optional[QCloseEvent]) -> None:
        self.widget.batch_widget.hide()
        if a0 is not None:
            a0.ignore()

    def set_3d_view_f(self):
        pg_layout = self.widget.batch_widget.surface_widget.surface_view.pg_layout
        pg_layout.opts["azimuth"] = 0
        pg_layout.opts["elevation"] = 0
        pg_layout.opts["distance"] = 3
        self.plot_batch()

    def set_3d_view_t(self):
        pg_layout = self.widget.batch_widget.surface_widget.surface_view.pg_layout
        pg_layout.opts["azimuth"] = 0
        pg_layout.opts["elevation"] = 90
        pg_layout.opts["distance"] = 3
        self.plot_batch()

    def set_3d_view_s(self):
        pg_layout = self.widget.batch_widget.surface_widget.surface_view.pg_layout
        pg_layout.opts["azimuth"] = 90
        pg_layout.opts["elevation"] = 0
        pg_layout.opts["distance"] = 3
        self.plot_batch()

    def set_3d_view_i(self):
        pg_layout = self.widget.batch_widget.surface_widget.surface_view.pg_layout
        pg_layout.opts["azimuth"] = 45
        pg_layout.opts["elevation"] = 30
        pg_layout.opts["distance"] = 3
        self.plot_batch()

    def pressed_button_x(self):
        self.key_pressed_3d(None, pressed_key=88)

    def pressed_button_y(self):
        self.key_pressed_3d(None, pressed_key=89)

    def pressed_button_z(self):
        self.key_pressed_3d(None, pressed_key=90)

    def pressed_button_1(self):
        self.key_pressed_3d(None, pressed_key=49)

    def pressed_button_2(self):
        self.key_pressed_3d(None, pressed_key=50)

    def pressed_button_3(self):
        self.key_pressed_3d(None, pressed_key=51)

    def pressed_button_h(self):
        self.key_pressed_3d(None, pressed_key=72)

    def pressed_button_l(self):
        self.key_pressed_3d(None, pressed_key=76)

    def pressed_button_g(self):
        self.key_pressed_3d(None, pressed_key=71)

    def pressed_button_m(self, ev):
        if ev.button() & QtCore.Qt.RightButton:
            val, ok = QtWidgets.QInputDialog.getInt(
                self.widget.integration_image_widget,
                "Edit marker size",
                "marker size:",
                min=0,
                max=25,
                value=self.widget.batch_widget.surface_widget.surface_view.marker_size,
            )
            if ok:
                self.widget.batch_widget.surface_widget.surface_view.marker_size = val
        self.key_pressed_3d(None, pressed_key=77)

    def pressed_button_s(self):
        self.key_pressed_3d(None, pressed_key=83)

    def set_marker_color(self):
        color = (
            self.widget.batch_widget.surface_widget.control_widget.m_color_btn.color(
                mode="float"
            )
        )
        self.widget.batch_widget.surface_widget.surface_view.marker_color = color[:3]
        self.plot_batch()

    def key_pressed_3d(self, ev, pressed_key=None):
        navigation_widget = self.widget.batch_widget.surface_widget.control_widget
        if pressed_key is None:
            pressed_key = ev.key()
        if pressed_key == 49:
            self.widget.batch_widget.scale_lin_btn.setChecked(True)
            self.scale = np.array
        if pressed_key == 50:
            self.widget.batch_widget.scale_sqrt_btn.setChecked(True)
            self.scale = np.sqrt
        if pressed_key == 51:
            self.widget.batch_widget.scale_log_btn.setChecked(True)
            self.scale = np.log10
        if pressed_key == 66:
            if self.widget.batch_widget.options_widget.background_btn.isChecked():
                self.widget.batch_widget.options_widget.background_btn.setChecked(False)
            else:
                self.widget.batch_widget.options_widget.background_btn.setChecked(True)
        if pressed_key == 72:
            navigation_widget.trim_h_btn.setChecked(True)
        if pressed_key == 76:
            navigation_widget.trim_l_btn.setChecked(True)
        if pressed_key == 88:
            navigation_widget.scale_x_btn.setChecked(True)
        if pressed_key == 89:
            navigation_widget.scale_y_btn.setChecked(True)
        if pressed_key == 90:
            navigation_widget.scale_z_btn.setChecked(True)
        if pressed_key == 83:
            navigation_widget.scale_s_btn.setChecked(True)
        if pressed_key == 71:
            navigation_widget.move_g_btn.setChecked(True)
        if pressed_key == 77:
            navigation_widget.move_m_btn.setChecked(True)
        if pressed_key == 70:
            self.widget.batch_widget.view3d_f_btn.setChecked(True)
            self.set_3d_view_f()
        if pressed_key == 82:
            self.widget.batch_widget.view3d_s_btn.setChecked(True)
            self.set_3d_view_s()
        if pressed_key == 84:
            self.widget.batch_widget.view3d_t_btn.setChecked(True)
            self.set_3d_view_t()
        if pressed_key == 73:
            self.widget.batch_widget.view3d_i_btn.setChecked(True)
            self.set_3d_view_i()
        if pressed_key in [72, 76, 88, 89, 90, 71, 77, 83]:
            self.widget.batch_widget.surface_widget.surface_view.pressed_key = (
                pressed_key
            )

        self.plot_batch()

    def wheel_event_3d(self, ev):

        data = self.model.batch_model.data
        binning = self.model.batch_model.binning

        layout = self.widget.batch_widget.surface_widget.pg_layout
        pressed_key = self.widget.batch_widget.surface_widget.surface_view.pressed_key
        show_range = self.widget.batch_widget.surface_widget.surface_view.show_range
        show_scale = self.widget.batch_widget.surface_widget.surface_view.show_scale
        surf_view = self.widget.batch_widget.surface_widget.surface_view

        delta = ev.angleDelta().x()
        if delta == 0:
            delta = ev.angleDelta().y()
        diff = delta / 10000

        if pressed_key == 72:
            if 1 > show_range[1] + diff > show_range[0]:
                show_range[1] += diff
        elif pressed_key == 76:
            if 0 < show_range[0] + diff < show_range[1]:
                show_range[0] += diff
        elif pressed_key == 88:
            show_scale[0] += diff
        elif pressed_key == 89:
            show_scale[1] += diff
        elif pressed_key == 90:
            show_scale[2] += diff
        elif pressed_key == 71:
            start = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
                )
            )
            stop = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
                )
            )
            surf_view.g_translate = min(
                stop, max(start, surf_view.g_translate + int(diff * data.shape[0] * 4))
            )
            y = int(surf_view.g_translate)
            self.widget.batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.x_pos_lbl.setText(
                f"Img: {int(y):.0f}"
            )
            self.load_single_image(int(surf_view.marker), y)
        elif pressed_key == 77:
            if 0 <= surf_view.marker + int(diff * data.shape[1]) < data.shape[1]:
                surf_view.marker += int(diff * data.shape[1])
                tth = binning[int(surf_view.marker)]
                self.widget.batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.y_pos_lbl.setText(
                    f"2θ:{tth:.1f}"
                )
                self.widget.pattern_widget.set_pos_line(tth)
        else:
            if ev.modifiers() == QtCore.Qt.ControlModifier:
                layout.opts["fov"] *= 0.999**delta
            else:
                layout.opts["distance"] *= 0.999**delta
            layout.update()

        self.plot_batch()

    def update_vertical_line_pos(self, tth):
        if self.model.batch_model.binning is None:
            return

        start_x, stop_x = (
            self.widget.batch_widget.stack_plot_widget.img_view.x_bin_range
        )
        binning = self.model.batch_model.binning[start_x:stop_x]
        pos = get_partial_index(binning, tth)

        if pos is None:
            self.widget.batch_widget.stack_plot_widget.img_view.deactivate_vertical_line()
            return

        self.widget.batch_widget.stack_plot_widget.img_view.activate_vertical_line()
        self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.setValue(pos)

    def set_unit_tth(self):
        """
        Set 2th_deg unit on batch plot

        Corresponding buttons on batch and pattern widgets are checked.
        """
        self.widget.batch_widget.options_widget.tth_btn.setChecked(True)
        self.widget.integration_pattern_widget.tth_btn.setChecked(True)
        self.model.current_configuration.integration_unit = "2th_deg"
        self.widget.batch_widget.stack_plot_widget.img_view.bottom_axis_cake.setLabel(
            "2θ", "°"
        )
        self.widget.batch_widget.stack_plot_widget.img_view.img_view_box.invertX(False)
        self.update_x_axis()
        if not self.model.calibration_model.is_calibrated:
            x = (
                self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
            )
            y = (
                self.widget.batch_widget.position_widget.step_series_widget.slider.value()
            )
            self.plot_pattern(int(x), int(y))

    def set_unit_q(self):
        """
        Set q_A^-1 unit on batch plot

        Corresponding buttons on batch and pattern widgets are checked.
        """
        self.widget.batch_widget.options_widget.q_btn.setChecked(True)
        self.widget.integration_pattern_widget.q_btn.setChecked(True)
        self.model.current_configuration.integration_unit = "q_A^-1"
        self.widget.batch_widget.stack_plot_widget.img_view.img_view_box.invertX(False)
        self.widget.batch_widget.stack_plot_widget.img_view.bottom_axis_cake.setLabel(
            "Q", "A<sup>-1</sup>"
        )
        self.update_x_axis()
        if not self.model.calibration_model.is_calibrated:
            x = (
                self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
            )
            y = (
                self.widget.batch_widget.position_widget.step_series_widget.slider.value()
            )
            self.plot_pattern(int(x), int(y))

    def set_unit_d(self):
        """
        Set d_A unit on batch plot

        Corresponding buttons on batch and pattern widgets are checked.
        """
        self.widget.batch_widget.options_widget.d_btn.setChecked(True)
        self.widget.integration_pattern_widget.d_btn.setChecked(True)
        self.model.current_configuration.integration_unit = "d_A"
        self.widget.batch_widget.stack_plot_widget.img_view.bottom_axis_cake.setLabel(
            "d", "A"
        )
        self.update_x_axis()
        if not self.model.calibration_model.is_calibrated:
            x = (
                self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
            )
            y = (
                self.widget.batch_widget.position_widget.step_series_widget.slider.value()
            )
            self.plot_pattern(int(x), int(y))

    def toggle_show_phases(self):
        """
        Show and hide phases
        """
        if self.widget.batch_widget.control_widget.phases_btn.isChecked():
            self.widget.batch_widget.stack_plot_widget.img_view.show_all_visible_cake_phases(
                self.widget.phase_widget.phase_show_cbs
            )
            self.widget.batch_widget.control_widget.phases_btn.setText("Hide Phases")
        else:
            self.widget.batch_widget.stack_plot_widget.img_view.hide_all_cake_phases()
            self.widget.batch_widget.control_widget.phases_btn.setText("Show Phases")

    def subtract_background(self):
        """
        Toddle background subtraction in batch image
        """
        data = self.model.batch_model.data
        bkg = self.model.batch_model.bkg
        if data is None:
            return

        if bkg is None:
            self.widget.show_error_msg(
                "Background is not yet calculated. Calculate background."
            )
            self.widget.batch_widget.options_widget.background_btn.setChecked(False)
            return

        if data.shape != bkg.shape:
            self.widget.show_error_msg(
                f"Shape of data {data.shape} and background {bkg.shape} are different."
                "Recalculate background."
            )
            self.widget.batch_widget.options_widget.background_btn.setChecked(False)
            return

        self.plot_batch()

    def extract_background(self):
        """
        Extract background from batch data
        """
        if self.model.batch_model.n_img is None:
            return

        progress_dialog = get_progress_dialog(
            "Extracting background",
            "Abort",
            self.model.batch_model.n_img,
            self.widget.batch_widget,
        )

        def callback_fn(current_index):
            if progress_dialog.wasCanceled():
                return False
            progress_dialog.setValue(current_index)
            QtWidgets.QApplication.processEvents()
            return ~progress_dialog.wasCanceled()

        parameters = (
            self.widget.integration_control_widget.background_control_widget.get_bkg_pattern_parameters()
        )
        self.model.batch_model.extract_background(parameters, callback_fn)
        progress_dialog.close()

    def set_hard_minimum(self, ev, scale):
        if ev.button() == QtCore.Qt.RightButton:
            val, ok = QtWidgets.QInputDialog.getDouble(
                self.widget.integration_image_widget,
                "Edit minimal value",
                "value:",
                decimals=3,
                value=self.min_val[scale],
            )
            if ok:
                self.min_val[scale] = val
        self.min_val["current"] = self.min_val[scale]

    def change_scale_log(self, e: typing.Optional[QMouseEvent]):
        """
        Change scale to log. Edit hard minimum of image value
        """
        self.widget.batch_widget.options_widget.scale_log_btn.setChecked(True)
        self.set_hard_minimum(e, "log")
        self.scale = np.log10
        self.plot_batch()

    def change_scale_lin(self, ev):
        """
        Change scale to linear. Edit hard minimum of image value
        """
        self.widget.batch_widget.options_widget.scale_lin_btn.setChecked(True)
        self.set_hard_minimum(ev, "lin")
        self.scale = np.array
        self.plot_batch()

    def change_scale_sqrt(self, ev):
        """
        Change scale to square root. Edit hard minimum of image value
        """
        self.widget.batch_widget.options_widget.scale_sqrt_btn.setChecked(True)
        self.set_hard_minimum(ev, "sqrt")
        self.scale = np.sqrt
        self.plot_batch()

    def process_step(self):
        """
        Re-draw waterfall plot if step value get changed.
        """
        if self.widget.batch_widget.control_widget.waterfall_btn.isChecked():
            self.plot_waterfall()
        if self.widget.batch_widget.mode_widget.view_3d_btn.isChecked():
            self.plot_batch()

    def waterfall_mode(self):
        """
        Set/unset widget in waterfall mode
        """
        if self.widget.batch_widget.control_widget.waterfall_btn.isChecked():
            self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.setVisible(
                False
            )
            self.widget.batch_widget.stack_plot_widget.img_view.horizontal_line.setVisible(
                False
            )
        else:
            if self.rect is not None:
                self.widget.batch_widget.stack_plot_widget.img_view.img_view_box.removeItem(
                    self.rect
                )
            self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.setVisible(
                True
            )
            self.widget.batch_widget.stack_plot_widget.img_view.horizontal_line.setVisible(
                True
            )

    def set_range_img(self):
        """
        Set start and stop value in the navigation widget
        """
        start = int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        stop = int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
            )
        )
        n_img_all = self.model.batch_model.n_img_all
        if n_img_all is None or stop < 0:
            return

        start = min(start, n_img_all, stop)
        stop = min(max(start, stop), n_img_all)

        self.widget.batch_widget.position_widget.step_series_widget.slider.setRange(
            start, stop
        )
        self.widget.batch_widget.position_widget.step_series_widget.pos_validator.setRange(
            start, stop
        )
        self.plot_batch()

    def switch_frame(self, y):
        x = self.widget.batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
        self.load_single_image(x, y)
        if self.widget.batch_widget.mode_widget.view_3d_btn.isChecked():
            self.widget.batch_widget.surface_widget.surface_view.g_translate = y
            self.plot_batch()

    def show_img_mouse_position(self, x, y):
        """
        Show position of the mouse with respect of the heatmap

        Show image number, position in diffraction pattern and intensity
        """
        img = self.model.batch_model.data
        binning = self.model.batch_model.binning
        y += int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )

        if img is None or x > img.shape[1] or x < 0 or y > img.shape[0] or y < 0:
            return
        scale = (binning[-1] - binning[0]) / binning.shape[0]
        tth = x * scale + binning[0]

        bkg = self.model.batch_model.bkg
        if (
            self.widget.batch_widget.options_widget.background_btn.isChecked()
            and bkg is not None
        ):
            z = img[int(y), int(x)] - bkg[int(y), int(x)]
        else:
            z = img[int(y), int(x)]

        self.widget.batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.x_pos_lbl.setText(
            f"Img: {int(y):.0f}"
        )
        self.widget.batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.y_pos_lbl.setText(
            f"2θ: {tth:.1f}"
        )
        self.widget.batch_widget.position_widget.mouse_pos_widget.cur_pos_widget.int_lbl.setText(
            f"I: {z:.1f}"
        )

    def change_view(self):
        """
        Change between 2D, 3D and file views
        """
        if self.widget.batch_widget.mode_widget.view_f_btn.isChecked():
            self.widget.batch_widget.activate_files_view()
            n_img_all = self.model.batch_model.n_img_all
            if n_img_all is not None:
                self.set_navigation_raw((0, n_img_all - 1))

        elif self.widget.batch_widget.mode_widget.view_3d_btn.isChecked():
            n_img = self.model.batch_model.n_img
            if n_img is None:
                self.widget.batch_widget.mode_widget.view_f_btn.setChecked(True)
                return
            self.set_navigation_range((0, n_img - 1))
            self.widget.batch_widget.activate_surface_view()
            self.plot_batch()
        else:
            n_img = self.model.batch_model.n_img
            if n_img is None:
                self.widget.batch_widget.mode_widget.view_f_btn.setChecked(True)
                return
            self.set_navigation_range((0, n_img - 1))
            self.widget.batch_widget.activate_stack_plot()

            self.plot_batch()

    def filename_txt_changed(self):
        """
        Set image files of the batch base on filename given in the text box
        """
        current_filenames = self.model.batch_model.files
        if current_filenames is None:
            return
        current_directory = self.model.working_directories["image"]

        img_filename_txt = str(self.widget.img_filename_txt.text())
        new_filenames = []
        for t in img_filename_txt.split():
            new_filenames += glob(os.path.join(current_directory, t))

        if len(new_filenames) > 0:
            try:
                self.load_raw_data(new_filenames)
            except TypeError:
                basenames = [os.path.basename(f) for f in current_filenames]
                self.widget.img_filename_txt.setText(" ".join(basenames))
        else:
            basenames = [os.path.basename(f) for f in current_filenames]
            self.widget.img_filename_txt.setText(" ".join(basenames))

    def directory_txt_changed(self):
        """
        Change directory name for image files of the batch
        """
        new_directory = str(self.widget.img_directory_txt.text())
        filenames = self.model.batch_model.files
        if filenames is None:
            return
        new_filenames = [
            os.path.join(new_directory, os.path.basename(f)) for f in filenames
        ]
        self.load_raw_data(new_filenames)

    def load_data(self):
        """
        Set image files of the batch, base on files given in the dialog window
        """
        start_dir = self.model.working_directories.get("batch", os.path.expanduser("~"))
        filenames = open_files_dialog(
            self.widget,
            "Load image data file(s)",
            start_dir,
            ("Raw data (*.nxs *.tif *.tiff *.h5);;" "Proc data (*.nxs)"),
        )
        if len(filenames) == 0:
            return
        else:
            self._load_data(filenames)

    def _load_data(self, filenames):
        self.model.working_directories["batch"] = os.path.dirname(filenames[0])
        self.widget.batch_widget.file_control_widget.folder_lbl.setText(
            os.path.dirname(filenames[0])
        )
        self.reset_view()
        if self.is_proc(filenames[0]):
            self.model.batch_model.reset_data()
            self.load_proc_data(filenames[0])
            self.load_raw_data(self.model.batch_model.files)
            self.widget.batch_widget.mode_widget.view_2d_btn.setChecked(True)
            self.change_view()
            self.plot_batch()
            self.widget.batch_widget.stack_plot_widget.img_view.auto_range()
        else:
            self.widget.img_directory_txt.setText(os.path.dirname(filenames[0]))
            self.model.batch_model.reset_data()
            self.load_raw_data(filenames)
            self.widget.batch_widget.mode_widget.view_f_btn.setChecked(True)
            self.change_view()

        self.load_single_image(1, 0)

        if self.model.calibration_model.is_calibrated:
            self.integrate()

    def load_previous_folder(self):
        filenames = self.model.batch_model.get_previous_folder_filenames()
        if len(filenames) > 0:
            self._load_data(filenames)

    def load_next_folder(self):
        filenames = self.model.batch_model.get_next_folder_filenames()
        if len(filenames) > 0:
            self._load_data(filenames)

    def reset_view(self):
        """
        Set few view buttons to un-checked.
        This brings batch-widget to initial state
        """
        self.widget.batch_widget.options_widget.background_btn.setChecked(False)
        self.widget.batch_widget.control_widget.waterfall_btn.setChecked(False)
        self.widget.batch_widget.control_widget.phases_btn.setChecked(False)

        self.waterfall_mode()
        self.toggle_show_phases()

    def is_proc(self, filename):
        """
        Check if file contains processed data
        """
        if os.path.splitext(filename)[1] == ".nxs":
            data_file = h5py.File(filename, "r")
            if "processed" in data_file:
                return True
            # ToDo Check for old format. To be removed
            if "data" in data_file:
                return True
        return False

    def load_raw_data(self, filenames):
        """
        Load metadata for raw data

        Following information is loaded:
            filenames, number of images in each file
        """

        self.model.img_model.blockSignals(True)
        try:
            self.model.batch_model.set_image_files(filenames)
        except IOError:
            try:
                dir_name = self.model.working_directories["image"]
                filenames = [
                    os.path.join(dir_name, os.path.basename(f)) for f in filenames
                ]
                self.model.batch_model.set_image_files(filenames)
            except IOError:
                print("Raw images are not found")

        self.model.img_model.blockSignals(False)
        files = self.model.batch_model.files
        file_map = self.model.batch_model.file_map

        if files is not None:
            if len(file_map) == len(files):
                file_map = np.hstack((file_map, self.model.batch_model.data.shape[0]))
            images = [(file_map[i + 1] - file_map[i]) for i in range(len(files))]
            self.widget.batch_widget.file_view_widget.set_raw_files(files, images)

        self.show_metadata_info()

        n_img = self.model.batch_model.n_img
        n_img_all = self.model.batch_model.n_img_all
        self.widget.batch_widget.position_widget.step_series_widget.pos_label.setText(
            f"Frame({n_img}/{n_img_all}):"
        )
        self.widget.batch_widget.position_widget.step_raw_widget.pos_label.setText(
            f"Frame({n_img}/{n_img_all}):"
        )

    def show_metadata_info(self):
        """
        Show calibration file and mask file in the file view
        """
        self.widget.batch_widget.file_view_widget.set_cal_file(
            self.model.batch_model.used_calibration
        )
        self.widget.batch_widget.file_view_widget.set_mask_file(
            self.model.batch_model.used_mask
        )

    def load_proc_data(self, filename):
        """
        Load processed data (diffraction patterns and metadata)
        """
        self.model.batch_model.load_proc_data(filename)
        self.widget.calibration_lbl.setText(
            self.model.calibration_model.calibration_name
        )

    def plot_batch(self, start=None, stop=None):
        """
        Plot batch of diffraction patters taking into account scale abd background subtraction
        """
        data = self.model.batch_model.data
        bkg = self.model.batch_model.bkg
        if data is None:
            return
        if (
            self.widget.batch_widget.options_widget.background_btn.isChecked()
            and bkg is not None
        ):
            data = data - bkg
        if self.min_val.get("current", None) is not None:
            data[data < self.min_val["current"]] = self.min_val["current"]
        data = self.scale(data)

        start_x, stop_x = self._get_x_range()
        if stop is None:
            stop = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.stop_txt.text()
                )
            )
        if start is None:
            start = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
                )
            )

        if self.widget.batch_widget.mode_widget.view_2d_btn.isChecked():
            self.widget.batch_widget.stack_plot_widget.img_view.plot_image(
                data[start : stop + 1, start_x:stop_x], True, [start_x, stop_x]
            )
            self.update_axes_range()
            self.update_linear_region()

        if self.widget.batch_widget.mode_widget.view_3d_btn.isChecked():
            step = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.step_txt.text()
                )
            )
            step_min = max(1, int(data[start : stop + 1].size / self.size_threshold))
            if step < step_min:
                step = step_min
                self.widget.batch_widget.position_widget.step_series_widget.step_txt.setValue(
                    step
                )
            self.widget.batch_widget.surface_widget.surface_view.plot_surface(
                data[start : stop + 1 : step, start_x:stop_x], start, step
            )
            self.update_3d_axis(data[start : stop + 1 : step, start_x:stop_x])

        self.model.enabled_phases_in_cake.emit()

    def _get_x_range(self):
        """
        Return bin-x range of the batch plot
        """
        if self.model.batch_model.data is None:
            return 0, 0
        start_x = 0
        stop_x = self.model.batch_model.data.shape[1]
        if self.widget.batch_widget.options_widget.bkg_cut_btn.isChecked():
            bkg_roi = self.model.pattern_model.pattern.auto_bkg_roi
            if bkg_roi is not None:
                bkg_roi = self.convert_x_value(
                    np.array(bkg_roi),
                    self.model.current_configuration.integration_unit,
                    "2th_deg",
                )
                binning = self.model.batch_model.binning
                scale = (binning[-1] - binning[0]) / binning.shape[0]
                start_x = (bkg_roi[0] - binning[0]) / scale
                stop_x = (bkg_roi[1] - binning[0]) / scale
        return int(start_x), int(stop_x)

    def update_linear_region(self):
        """
        Update linear region of 2D-view to background roi
        """
        bkg_roi = self.model.pattern_model.pattern.auto_bkg_roi
        if self.model.batch_model.binning is not None and bkg_roi is not None:
            bkg_roi = self.convert_x_value(
                np.array(bkg_roi),
                self.model.current_configuration.integration_unit,
                "2th_deg",
            )
            start_x, stop_x = (
                self.widget.batch_widget.stack_plot_widget.img_view.x_bin_range
            )
            binning = self.model.batch_model.binning[start_x:stop_x]
            scale = (binning[-1] - binning[0]) / binning.shape[0]
            x_min_bin = (bkg_roi[0] - binning[0]) / scale
            x_max_bin = (bkg_roi[1] - binning[0]) / scale
            self.widget.batch_widget.stack_plot_widget.img_view.set_linear_region(
                x_min_bin, x_max_bin
            )

    def save_data(self):
        """
        Save diffraction patterns and metadata
        """
        filename = save_file_dialog(
            self.widget,
            "Save Image.",
            directory=os.path.join(
                self.model.working_directories.get("batch", os.path.expanduser("~"))
            ),
            filter=(
                "Image (*.png);;Single file ascii (*.csv);;"
                "Multifile Data (*.xy);;"
                "Multifile Data (*.chi);;"
                "Multifile Data (*.dat);;"
                "Multifile GSAS (*.fxye);;"
                "Single file Data (*.nxs)"
            ),
        )

        name, ext = os.path.splitext(filename)
        if filename != "":
            if ext == ".png":
                if self.widget.batch_widget.mode_widget.view_2d_btn.isChecked():
                    self.widget.batch_widget.stack_plot_widget.img_view.save_img(
                        filename
                    )
                if self.widget.batch_widget.mode_widget.view_3d_btn.isChecked():
                    d = self.widget.batch_widget.surface_widget.surface_view.pg_layout.renderToArray(
                        (1000, 1000)
                    )
                    makeQImage(d).save(filename)
            elif ext == ".nxs":
                self.model.batch_model.save_proc_data(filename)
            elif ext == ".csv":
                self.model.batch_model.save_as_csv(filename)
            else:
                self.model.img_model.blockSignals(True)
                img_data = self.model.batch_model.data
                pattern_x = self.model.batch_model.binning
                for y in range(img_data.shape[0]):
                    pattern_y = img_data[int(y)]
                    self.model.pattern_model.set_pattern(pattern_x, pattern_y)
                    self.model.current_configuration.save_pattern(
                        f"{name}_{y:03d}{ext}"
                    )

                    if self.model.pattern_model.pattern.auto_bkg is not None:
                        bkg_directory = os.path.join(
                            os.path.dirname(filename), "bkg_subtracted"
                        )
                        if not os.path.exists(bkg_directory):
                            os.mkdir(bkg_directory)
                        bkg_subtracted_name = os.path.join(
                            bkg_directory, os.path.basename(name)
                        )

                        self.model.pattern_model.pattern.recalculate_pattern()
                        self.model.current_configuration.save_pattern(
                            f"{bkg_subtracted_name}_{y:03d}{ext}",
                            subtract_background=True,
                        )
                self.model.img_model.blockSignals(False)

    def img_mouse_click(self, x, y):
        """
        Process mouse click
        """
        y += int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        if self.widget.batch_widget.control_widget.waterfall_btn.isChecked():
            self.process_waterfall(x, y)
        else:

            start_x, stop_x = (
                self.widget.batch_widget.stack_plot_widget.img_view.x_bin_range
            )
            binning = self.model.batch_model.binning[start_x:stop_x]
            tth = get_partial_value(binning, x - 0.5)
            if tth is not None:
                self.model.clicked_tth_changed.emit(tth)
                self.load_single_image(x, y)

    def img_autoscale_btn_clicked(self):
        self.widget.batch_widget.stack_plot_widget.img_view.auto_level()

    def normalize_btn_clicked(self):
        self.model.batch_model.normalize()
        self.plot_batch()

    def process_waterfall(self, x, y):
        """
        Create waterfall plot based on selected rectangle in the heatmap
        """
        # show overlay widget
        self.widget.integration_control_widget.tab_widget_1.setCurrentWidget(
            self.widget.integration_control_widget.overlay_control_widget
        )
        self.widget.integration_control_widget.tab_widget_2.setCurrentWidget(
            self.widget.integration_control_widget.overlay_control_widget
        )

        if self.clicks == 0:
            self.clicks += 1
            if self.rect is not None:
                self.widget.batch_widget.stack_plot_widget.img_view.img_view_box.removeItem(
                    self.rect
                )
            start = int(
                str(
                    self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
                )
            )
            self.rect = (
                self.widget.batch_widget.stack_plot_widget.img_view.draw_rectangle(
                    x, y - start
                )
            )
            self.widget.batch_widget.stack_plot_widget.img_view.mouse_moved.connect(
                self.rect.set_size
            )
            self.plot_pattern(int(x), int(y))
        elif self.clicks == 1:
            self.clicks = 0
            self.widget.batch_widget.stack_plot_widget.img_view.mouse_moved.disconnect(
                self.rect.set_size
            )
            self.plot_waterfall()

    def plot_waterfall(self):
        """
        Create waterfall plot based on position and size of rectangle
        """
        data = self.model.batch_model.data
        start_x, stop_x = self._get_x_range()
        start = int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        binning = self.model.batch_model.binning
        bkg = self.model.batch_model.bkg
        if data is None:
            return
        if self.widget.batch_widget.options_widget.background_btn.isChecked():
            data = data - bkg

        rect = self.rect.rect()
        y1, y2 = sorted((int(rect.top()), int(rect.bottom())))
        x1, x2 = sorted((int(rect.left()), int(rect.right())))
        y1 += start
        y2 += start
        x1 += start_x
        x2 += start_x

        y1 = max(y1, 0)
        y2 = min(y2, data.shape[0])
        x1 = max(x1, start_x)
        x2 = min(x2, stop_x)

        step = int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.step_txt.text()
            )
        )
        self.model.overlay_model.reset()
        new_binning = self.convert_x_value(
            binning[x1:x2], "2th_deg", self.model.current_configuration.integration_unit
        )
        for i in range(y1, y2, step):
            f_name, pos = self.model.batch_model.get_image_info(i)
            f_name = os.path.basename(f_name)
            self.model.overlay_model.add_overlay(
                new_binning, data[i, x1:x2], f"{f_name}, {pos}"
            )
        separation = (
            self.widget.integration_control_widget.overlay_control_widget.waterfall_separation_msb.value()
        )
        self.model.overlay_model.overlay_waterfall(separation)

    def load_single_image(self, x, y):
        """
        Plot raw image, diffraction pattern and draw lines on the heatmap plot based on given x and y
        """
        img = self.model.batch_model.data
        if img is None:
            return
        x = min(max(x, 0), img.shape[1])
        y = min(max(y, 0), img.shape[0] - 1)
        self.widget.batch_widget.position_widget.step_series_widget.slider.setValue(
            int(y)
        )
        self.widget.batch_widget.position_widget.step_series_widget.pos_txt.setText(
            str(int(y))
        )
        self.plot_image(int(y))
        self.plot_pattern(int(x), int(y))
        start = int(
            str(
                self.widget.batch_widget.position_widget.step_series_widget.start_txt.text()
            )
        )
        self.widget.batch_widget.stack_plot_widget.img_view.horizontal_line.setValue(
            y - start
        )

    def plot_pattern(self, x, y):
        """
        Plot diffraction pattern using proc data
        """
        img = self.model.batch_model.data
        binning = self.model.batch_model.binning
        if (
            img is None
            or x > img.shape[1] - 1
            or x < 0
            or y > img.shape[0] - 1
            or y < 0
        ):
            return
        scale = (binning[-1] - binning[0]) / binning.shape[0]
        tth = x * scale + binning[0]
        z = img[y, x]

        self.widget.batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.y_pos_lbl.setText(
            f"2θ: {tth:.1f}"
        )
        self.widget.batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.int_lbl.setText(
            f"I: {z:.1f}"
        )
        new_binning = self.convert_x_value(
            binning, "2th_deg", self.model.current_configuration.integration_unit
        )
        self.model.pattern_model.set_pattern(new_binning, img[y])

    def plot_image(self, y):
        """
        Plot single raw image from the batch

        :param y: Number of raw image in the batch
        """
        y = int(y)
        self.model.current_configuration.auto_integrate_pattern = False
        self.model.batch_model.load_image(
            y, self.widget.batch_widget.mode_widget.view_f_btn.isChecked()
        )
        f_name, pos = self.model.batch_model.get_image_info(
            y, self.widget.batch_widget.mode_widget.view_f_btn.isChecked()
        )
        self.widget.batch_widget.setWindowTitle(f"Batch widget. {f_name} - {pos}")
        self.model.current_configuration.auto_integrate_pattern = True
        self.widget.batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.setText(
            f"Img: {y:.0f}"
        )

    def update_axes_range(self):
        """
        Update axis of the 2D image
        """
        self.update_x_axis()
        self.update_y_axis()

    def update_3d_axis(self, data):
        """
        Update axis and grid in 3D view
        """
        if self.model.batch_model.binning is None:
            return

        surf_view = self.widget.batch_widget.surface_widget.surface_view
        start_x, stop_x = (
            self.widget.batch_widget.stack_plot_widget.img_view.x_bin_range
        )
        binning = self.model.batch_model.binning[start_x:stop_x]

        size = surf_view.pg_layout.pixelSize(np.array([0, 0, 0]))
        space = round(size * binning.shape[0] * 0.3, 2)
        ticks = np.round(np.arange(binning[0], binning[-1], space), 2)
        if ticks.shape[0] < 2:
            ticks = np.array([binning[0], binning[-1]])
        # surf_view.axis.set_tick_values(yticks=np.round(np.arange(binning[0], binning[-1], space), 2))
        surf_view.back_grid.setSpacing(
            np.nanmax(data) / 8.0, data.shape[1] / (ticks.shape[0] - 1), 1
        )
        surf_view.base_grid.setSpacing(1, data.shape[1] / (ticks.shape[0] - 1), 1)

    def update_x_axis(self):
        if self.model.batch_model.binning is None:
            return

        img_view_rect = (
            self.widget.batch_widget.stack_plot_widget.img_view.img_view_rect()
        )
        start_x, stop_x = (
            self.widget.batch_widget.stack_plot_widget.img_view.x_bin_range
        )
        binning = self.model.batch_model.binning[start_x:stop_x]

        width = img_view_rect.width()
        left = img_view_rect.left()

        h_scale = (np.max(binning) - np.min(binning)) / binning.shape[0]
        h_shift = binning[0]
        min_tth = h_scale * left + h_shift
        max_tth = h_scale * (left + width) + h_shift

        if self.model.current_configuration.integration_unit == "q_A^-1":
            ticks = [self.get_ticks(max_tth, min_tth, "q_A^-1", "2th_deg")]
        elif self.model.current_configuration.integration_unit == "d_A":
            ticks = [self.get_ticks(min_tth, max_tth, "d_A", "2th_deg")]
        else:
            ticks = None

        self.widget.batch_widget.stack_plot_widget.img_view.bottom_axis_cake.setRange(
            min_tth, max_tth
        )
        self.widget.batch_widget.stack_plot_widget.img_view.bottom_axis_cake.setTicks(
            ticks
        )

    def get_ticks(self, min_val, max_val, ticks_unit, base_unit, n_ticks=8):
        """
        Calculate ticks for x axis in case of non-linear scale
        """
        max_d = self.convert_x_value(max_val, base_unit, ticks_unit)
        min_d = self.convert_x_value(min_val, base_unit, ticks_unit)
        step = (max_d - min_d) / n_ticks
        x_d = min_d
        ticks = []
        while x_d > max_d:
            x_d = round(x_d + step, abs(round(np.log10(abs(step)))) + 1)
            x_tth = self.convert_x_value(x_d, ticks_unit, base_unit)
            ticks.append((x_tth, f"{x_d}"))
        return ticks

    def convert_x_value(self, value, previous_unit, new_unit):
        wavelength = self.model.calibration_model.wavelength
        if previous_unit == "2th_deg":
            tth = value
        elif previous_unit == "q_A^-1":
            tth = np.arcsin(value * 1e10 * wavelength / (4 * np.pi)) * 360 / np.pi
        elif previous_unit == "d_A":
            tth = 2 * np.arcsin(wavelength / (2 * value * 1e-10)) * 180 / np.pi
        else:
            tth = 0

        if new_unit == "2th_deg":
            res = tth
        elif new_unit == "q_A^-1":
            res = 4 * np.pi * np.sin(tth / 360 * np.pi) / wavelength / 1e10
        elif new_unit == "d_A":
            res = wavelength / (2 * np.sin(tth / 360 * np.pi)) * 1e10
        else:
            res = 0
        return res

    def update_y_axis(self):
        """
        Update y-axis of the batch plot
        """
        if self.model.batch_model.data is None:
            return

        y = self.widget.batch_widget.position_widget.step_series_widget.slider.value()
        start, stop, step = (
            self.widget.batch_widget.position_widget.step_series_widget.get_image_range()
        )
        self.widget.batch_widget.stack_plot_widget.img_view.horizontal_line.setValue(
            y - start
        )

        img_view_box = self.widget.batch_widget.stack_plot_widget.img_view.img_view_box
        data_img_item = (
            self.widget.batch_widget.stack_plot_widget.img_view.data_img_item
        )
        img_data = self.model.batch_model.data[start : stop + 1]

        height = img_view_box.viewRect().height()
        bottom = img_view_box.viewRect().top()
        bound = data_img_item.boundingRect().height()

        if bound == 0:
            return
        v_scale = img_data.shape[0] / bound
        min_y = v_scale * bottom + start
        max_y = v_scale * (bottom + height) + start

        self.widget.batch_widget.stack_plot_widget.img_view.left_axis_cake.setRange(
            min_y, max_y
        )

    def integrate(self):
        """
        Integrate images in the batch
        """
        if not self.model.calibration_model.is_calibrated:
            self.widget.show_error_msg(
                "Can not integrate multiple images without calibration."
            )
            return
        if (
            not self.model.batch_model.raw_available
            or self.model.batch_model.n_img_all < 1
        ):
            self.widget.show_error_msg("No images loaded for integration")
            return

        if self.widget.batch_widget.mode_widget.view_f_btn.isChecked():
            start, stop, step = (
                self.widget.batch_widget.position_widget.step_raw_widget.get_image_range()
            )
        else:
            start, stop, step = (
                self.widget.batch_widget.position_widget.step_series_widget.get_image_range()
            )

        n_int = (stop - start) / step
        progress_dialog = get_progress_dialog(
            "Integrating multiple images.",
            "Abort Integration",
            n_int,
            self.widget.batch_widget,
        )

        def callback_fn(current_index):
            if progress_dialog.wasCanceled():
                return False
            progress_dialog.setValue(current_index)
            QtWidgets.QApplication.processEvents()
            return ~progress_dialog.wasCanceled()

        self.model.batch_model.integrate_raw_data(
            start,
            stop + 1,
            step,
            self.widget.batch_widget.mode_widget.view_f_btn.isChecked(),
            callback_fn=callback_fn,
        )

        progress_dialog.close()
        self.show_metadata_info()

        n_img = self.model.batch_model.n_img
        n_img_all = self.model.batch_model.n_img_all
        self.widget.batch_widget.position_widget.step_series_widget.pos_label.setText(
            f"Frame({n_img}/{n_img_all}):"
        )
        self.widget.batch_widget.position_widget.step_raw_widget.pos_label.setText(
            f"Frame({n_img}/{n_img_all}):"
        )
        self.widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(
            n_img - 1
        )
        self.widget.batch_widget.position_widget.step_series_widget.start_txt.setValue(
            0
        )
        self.widget.batch_widget.mode_widget.view_2d_btn.setChecked(True)
        self.reset_view()
        self.change_view()
        self.widget.batch_widget.stack_plot_widget.img_view.auto_range()

    def set_navigation_raw(self, raw_range=(0, 0)):
        self.widget.batch_widget.position_widget.step_raw_widget.start_txt.setRange(
            *raw_range
        )
        self.widget.batch_widget.position_widget.step_raw_widget.stop_txt.setRange(
            *raw_range
        )
        self.widget.batch_widget.position_widget.step_raw_widget.stop_txt.setValue(
            raw_range[1]
        )
        self.widget.batch_widget.position_widget.step_raw_widget.slider.setRange(
            *raw_range
        )
        self.widget.batch_widget.position_widget.step_raw_widget.pos_validator.setRange(
            *raw_range
        )

    def set_navigation_range(self, all_range):
        """
        Set start and stop positions as well as range of navigation widget
        """
        self.widget.batch_widget.position_widget.step_series_widget.start_txt.setRange(
            *all_range
        )
        self.widget.batch_widget.position_widget.step_series_widget.stop_txt.setRange(
            *all_range
        )

        self.widget.batch_widget.position_widget.step_series_widget.slider.setRange(
            *all_range
        )
        self.widget.batch_widget.position_widget.step_series_widget.pos_validator.setRange(
            *all_range
        )

        self.widget.batch_widget.position_widget.step_series_widget.start_txt.setValue(
            all_range[0]
        )
        self.widget.batch_widget.position_widget.step_series_widget.stop_txt.setValue(
            all_range[1]
        )

    def update_gui(self):
        """
        Apply integration unit from current_configuration
        """
        if self.model.current_configuration.integration_unit == "2th_deg":
            self.widget.batch_widget.options_widget.tth_btn.setChecked(True)
            self.set_unit_tth()
        elif self.model.current_configuration.integration_unit == "d_A":
            self.widget.batch_widget.options_widget.d_btn.setChecked(True)
            self.set_unit_d()
        elif self.model.current_configuration.integration_unit == "q_A^-1":
            self.widget.batch_widget.options_widget.q_btn.setChecked(True)
            self.set_unit_q()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial

from qtpy import QtWidgets, QtCore

from .CustomWidgets import (
    LabelAlignRight,
    HorizontalSpacerItem,
    CheckableButton,
    NumberTextField,
    IntegerTextField,
    VerticalLine,
    SaveIconButton,
)


class ConfigurationWidget(QtWidgets.QWidget):
    configuration_selected = QtCore.Signal(int)  # configuration index

    def __init__(self, parent=None):
        super(ConfigurationWidget, self).__init__(parent)
        self.btn_size = QtCore.QSize(20, 20)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()
        self.add_tooltips()

    def create_widgets(self):
        self.configuration_lbl = LabelAlignRight("Configuration:")

        self.configuration_btns = []
        self.configurations_btn_widget = QtWidgets.QWidget()
        self.configuration_btn_group = QtWidgets.QButtonGroup()

        self.add_configuration_btn = QtWidgets.QPushButton("+")
        self.remove_configuration_btn = QtWidgets.QPushButton("-")

        self.factor_lbl = LabelAlignRight("Factor: ")
        self.factor_txt = NumberTextField("1")

        self.file_lbl = LabelAlignRight("File: ")
        self.previous_file_btn = QtWidgets.QPushButton("<")
        self.next_file_btn = QtWidgets.QPushButton(">")
        self.file_iterator_pos_lbl = LabelAlignRight(" Pos: ")
        self.file_iterator_pos_txt = IntegerTextField("1")

        self.folder_lbl = LabelAlignRight(" Folder:")
        self.next_folder_btn = QtWidgets.QPushButton(">")
        self.previous_folder_btn = QtWidgets.QPushButton("<")
        self.mec_cb = QtWidgets.QCheckBox("MEC")

        self.combine_patterns_btn = CheckableButton("Combine Patterns")
        self.combine_cakes_btn = CheckableButton("Combine Cakes")
        self.saved_combined_patterns_btn = SaveIconButton()
        self.saved_combined_patterns_btn.setToolTip("Save combined pattern")

    def create_layout(self):
        self.main_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addWidget(self.configuration_lbl)
        self.main_layout.addWidget(self.add_configuration_btn)
        self.main_layout.addWidget(self.remove_configuration_btn)
        self.main_layout.addWidget(self.configurations_btn_widget)
        self.main_layout.addSpacerItem(HorizontalSpacerItem())
        self.main_layout.addWidget(self.file_lbl)
        self.main_layout.addWidget(self.previous_file_btn)
        self.main_layout.addWidget(self.next_file_btn)
        self.main_layout.addWidget(self.file_iterator_pos_lbl)
        self.main_layout.addWidget(self.file_iterator_pos_txt)
        self.main_layout.addWidget(VerticalLine())
        self.main_layout.addWidget(self.folder_lbl)
        self.main_layout.addWidget(self.previous_folder_btn)
        self.main_layout.addWidget(self.next_folder_btn)
        self.main_layout.addWidget(self.mec_cb)
        self.main_layout.addSpacerItem(HorizontalSpacerItem())
        self.main_layout.addWidget(self.factor_lbl)
        self.main_layout.addWidget(self.factor_txt)
        self.main_layout.addSpacerItem(
            QtWidgets.QSpacerItem(
                20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
            )
        )
        self.main_layout.addWidget(self.combine_patterns_btn)
        self.main_layout.addWidget(self.saved_combined_patterns_btn)
        self.main_layout.addWidget(self.combine_cakes_btn)
        self.setLayout(self.main_layout)

        self.configurations_btn_layout = QtWidgets.QHBoxLayout(
            self.configurations_btn_widget
        )

    def style_widgets(self):
        self.main_layout.setSpacing(6)
        self.main_layout.setContentsMargins(6, 0, 6, 0)
        self.configurations_btn_layout.setSpacing(3)
        self.configurations_btn_layout.setContentsMargins(0, 0, 0, 0)

        btns = [
            self.add_configuration_btn,
            self.remove_configuration_btn,
            self.next_file_btn,
            self.previous_file_btn,
            self.next_folder_btn,
            self.previous_folder_btn,
            self.saved_combined_patterns_btn,
        ]

        for btn in btns:
            btn.setFixedSize(self.btn_size)

        self.saved_combined_patterns_btn.setIconSize(QtCore.QSize(13, 13))
        self.saved_combined_patterns_btn.setToolTip("Save combined pattern")
        self.saved_combined_patterns_btn.setFixedSize(25, 25)

        self.combine_cakes_btn.setFixedHeight(25)
        self.combine_patterns_btn.setFixedHeight(25)


    def update_configuration_btns(self, configurations, cur_ind):
        for btn in self.configuration_btns:
            self.configuration_btn_group.removeButton(btn)
            self.configurations_btn_layout.removeWidget(btn)
            btn.deleteLater()  # somehow needs tobe deleted, otherwise remains in the button group

        self.configuration_btns = []

        for ind, configuration in enumerate(configurations):
            new_button = CheckableButton(str(ind + 1))
            new_button.setFixedSize(25, 25)
            new_button.setToolTip("Switch to configuration {}".format(ind + 1))
            self.configuration_btn_group.addButton(new_button)
            self.configuration_btns.append(new_button)
            self.configurations_btn_layout.addWidget(new_button)
            if ind == cur_ind:
                new_button.setChecked(True)
            new_button.clicked.connect(partial(self.configuration_selected.emit, ind))

    def add_tooltips(self):
        self.add_configuration_btn.setToolTip("Add configuration")
        self.remove_configuration_btn.setToolTip("Remove configuration")
        self.next_file_btn.setToolTip("Next file")
        self.previous_file_btn.setToolTip("Previous file")
        self.next_folder_btn.setToolTip("Next folder")
        self.previous_folder_btn.setToolTip("Previous folder")



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from qtpy import QtWidgets, QtGui, QtCore

from .ConfigurationWidget import ConfigurationWidget
from .CalibrationWidget import CalibrationWidget
from .MaskWidget import MaskWidget
from .integration import IntegrationWidget
from .MapWidget import MapWidget
from .CustomWidgets import (
    VerticalSpacerItem,
    CheckableFlatButton,
    FlatButton,
    VerticalLine,
    HorizontalLine,
)

from .. import icons_path


class MainWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MainWidget, self).__init__(*args, **kwargs)

        self._create_layouts()

        self.setLayout(self._outer_layout)

        self._create_menu()
        self._create_mode_menu()

        self._left_layout.addLayout(self._menu_layout)
        self._left_layout.addLayout(self._mode_layout)
        self._left_layout.addLayout(self._external_actions_layout)

        self._outer_layout.addLayout(self._left_layout)
        self._outer_layout.addWidget(VerticalLine())
        self._outer_layout.addLayout(self._content_layout)

        self.configuration_widget = ConfigurationWidget(self)
        self.configuration_widget.setVisible(False)
        self._content_layout.addWidget(self.configuration_widget)

        self._create_main_frame()

        self._style_layouts()
        self.style_widgets()
        self.add_menu_popup()
        self.add_tooltips()

        self._content_layout.addWidget(self.main_frame)
        self.setWindowIcon(QtGui.QIcon(os.path.join(icons_path, "icon.svg")))

    def _style_layouts(self):
        self._outer_layout.setContentsMargins(0, 0, 0, 0)
        self._outer_layout.setStretchFactor(self._left_layout, 0)
        self._outer_layout.setStretchFactor(self._content_layout, 100)
        self._outer_layout.setSpacing(0)
        self._mode_layout.setContentsMargins(0, 0, 0, 0)
        self._mode_layout.setSpacing(0)
        self._content_layout.setSpacing(0)
        self._content_layout.setContentsMargins(0, 6, 0, 0)
        self._left_layout.setContentsMargins(0, 0, 0, 0)
        self._left_layout.setSpacing(6)
        self._external_actions_layout.setContentsMargins(6, 6, 6, 6)
        self._external_actions_layout.setSpacing(6)

    def _create_layouts(self):
        self._outer_layout = QtWidgets.QHBoxLayout()
        self._left_layout = QtWidgets.QVBoxLayout()
        self._external_actions_layout = QtWidgets.QVBoxLayout()
        self._content_layout = QtWidgets.QVBoxLayout()
        self._mode_layout = QtWidgets.QVBoxLayout()

    def _create_mode_menu(self):
        self.mode_btn_group = QtWidgets.QButtonGroup()
        self.calibration_mode_btn = CheckableFlatButton("CALIB", self)
        self.calibration_mode_btn.setObjectName("calibration_mode_btn")
        self.calibration_mode_btn.setChecked(True)
        self.mask_mode_btn = CheckableFlatButton("MASK", self)
        self.mask_mode_btn.setObjectName("mask_mode_btn")
        self.integration_mode_btn = CheckableFlatButton("INT", self)
        self.integration_mode_btn.setObjectName("integration_mode_btn")
        self.map_mode_btn = CheckableFlatButton("MAP", self)
        self.map_mode_btn.setObjectName("map_mode_btn")

        self.mode_btn_group.addButton(self.calibration_mode_btn)
        self.mode_btn_group.addButton(self.mask_mode_btn)
        self.mode_btn_group.addButton(self.integration_mode_btn)
        self.mode_btn_group.addButton(self.map_mode_btn)

        self._mode_layout.addWidget(self.calibration_mode_btn)
        self._mode_layout.addWidget(HorizontalLine())
        self._mode_layout.addWidget(self.mask_mode_btn)
        self._mode_layout.addWidget(HorizontalLine())
        self._mode_layout.addWidget(self.integration_mode_btn)
        self._mode_layout.addWidget(HorizontalLine())
        self._mode_layout.addWidget(self.map_mode_btn)
        self._mode_layout.addSpacerItem(VerticalSpacerItem())

    def _create_menu(self):
        self.menu_btn = QtWidgets.QPushButton("...")
        self.menu_btn.setObjectName("menu_btn")

        self.show_configuration_menu_btn = CheckableFlatButton("C")

        self._menu_layout = QtWidgets.QHBoxLayout()
        self._menu_layout.setContentsMargins(6, 6, 3, 0)
        self._menu_layout.setSpacing(12)

        self._menu_layout.addWidget(self.menu_btn)
        self._menu_layout.addWidget(self.show_configuration_menu_btn)

        self.save_btn = FlatButton("Save Project")
        self.load_btn = FlatButton("Open Project")
        self.reset_btn = FlatButton("Reset Project")

    def _create_main_frame(self):
        self.main_frame = QtWidgets.QWidget(self)
        self._layout_main_frame = QtWidgets.QVBoxLayout()
        self._layout_main_frame.setContentsMargins(0, 10, 6, 6)
        self._layout_main_frame.setSpacing(0)
        self.main_frame.setLayout(self._layout_main_frame)

        self.calibration_widget = CalibrationWidget(self)
        self.mask_widget = MaskWidget(self)
        self.integration_widget = IntegrationWidget(self)
        self.map_widget = MapWidget(self)

        self._layout_main_frame.addWidget(self.calibration_widget)
        self._layout_main_frame.addWidget(self.mask_widget)
        self._layout_main_frame.addWidget(self.integration_widget)
        self._layout_main_frame.addWidget(self.map_widget)

        self.mask_widget.setVisible(False)
        self.integration_widget.setVisible(False)
        self.map_widget.setVisible(False)

        self._content_layout.addWidget(self.main_frame)
        self._content_layout.setStretchFactor(self.main_frame, 100)

        self.style_widgets()
        self.add_tooltips()

        self.setWindowIcon(QtGui.QIcon(os.path.join(icons_path, "icon.svg")))

    def style_widgets(self):
        self._style_mode_btns()
        self._style_menu_btn()

        button_height = 24
        button_width = 24
        adjust_height_btns = [
            self.show_configuration_menu_btn,
        ]
        for btn in adjust_height_btns:
            btn.setHeight(button_height)
            btn.setWidth(button_width)

        self.configuration_widget.setMaximumHeight(28)

    def _style_menu_btn(self):
        self.menu_btn.setFixedWidth(30)
        self.menu_btn.setFixedHeight(30)

    def _style_mode_btns(self):
        mode_btn_width = 75
        mode_btn_height = 75
        mode_btns = [
            self.calibration_mode_btn,
            self.mask_mode_btn,
            self.integration_mode_btn,
            self.map_mode_btn,
        ]
        for btn in mode_btns:
            # btn.setCheckable(True)
            btn.setFixedWidth(mode_btn_width)
            btn.setFixedHeight(mode_btn_height)

    def add_menu_popup(self):
        self.menu_btn.clicked.connect(self.show_menu_popup)

    def show_menu_popup(self):
        widget = MenuPopup(self, [self.load_btn, self.save_btn, self.reset_btn])
        btn = self.menu_btn
        widget.adjustSize()
        position = self.mapToGlobal(QtCore.QPoint(btn.x() + btn.width() + 3, btn.y()))
        widget.move(position)
        widget.show()

    def add_tooltips(self):
        self.menu_btn.setToolTip("Project Menu")
        self.show_configuration_menu_btn.setToolTip("Show Configurations")
        self.calibration_mode_btn.setToolTip("Calibration Mode")
        self.mask_mode_btn.setToolTip("Mask Mode")
        self.integration_mode_btn.setToolTip("Integration Mode")

    def create_external_actions(self, quick_actions):
        self.external_action_btns = {}
        for action in quick_actions:
            btn = QtWidgets.QPushButton(action['name'])
            self.external_action_btns[action['name']] = btn
            self._external_actions_layout.addWidget(btn)


class MenuPopup(QtWidgets.QFrame):
    def __init__(self, parent=None, menu_items=None):
        super(MenuPopup, self).__init__(parent)
        self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.setFrameShadow(QtWidgets.QFrame.Raised)
        self.setLineWidth(1)
        self.setMidLineWidth(0)
        self.setObjectName("MenuPopup")
        self.setWindowFlags(QtCore.Qt.Popup | QtCore.Qt.FramelessWindowHint)
        self.setMouseTracking(True)
        self.setAcceptDrops(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.setWindowOpacity(0.9)
        self.setFixedWidth(150)
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)

        self.create_menu(menu_items)

    def create_menu(self, menu_items):
        for item in menu_items:
            self._layout.addWidget(item)
